#
# Collective Knowledge (analysis of variation of experimental results)
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin, Grigori.Fursin@cTuning.org, http://fursin.net
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}

##############################################################################
# analyze variation of experimental results

def analyze(i):
    """
    Input:  {
              characteristics_table - characteristics table (list)

              (bins)                - number of bins (int, default = 100)

              (min)                 - min float value (if multiple ctables are processed)
              (max)                 - max float value (if multiple ctables are processed)

              (cov_factor)          - float covariance factor (0.5 by default)

              (skip_fail)           - if 'yes', do not fail, if SciPy and NumPy
                                      are not available
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              min        - min X values from input or calculated
              max        - max X values from input or calculated  

              xlist      - list of x values
              ylist      - list of y density values 

              xlist2     - list of x values with peaks
              ylist2     - list of y density values with peaks

              xlist2s    - list of sorted x values with peaks (max y -> hence 1st expected value)
              ylist2s    - list of sorted y density values with peaks
            }

    """

    import copy

    has_deps=True
    try:
       from scipy.stats import gaussian_kde
       from scipy.signal import argrelextrema
       import numpy as np
    except Exception as e: 
       has_deps=False
       if i.get('skip_fail','')!='yes':
          return {'return':1, 'error':'Seems that some scientific python modules are not installed ('+format(e)+')'}

    ctable1=i['characteristics_table']
    ctable=copy.deepcopy(ctable1) # since slightly changing it ...

    dmin=i.get('min',-1)
    dmax=i.get('max',-1)

    xlist=[]
    ylist=[]

    xlist2=[]
    ylist2=[]

    xlist2s=[]
    ylist2s=[]


    dmin=i.get('min',-1.0)
    dmax=i.get('max',-1.0)

    if has_deps:
       if len(ctable)>0:
          xlistx=[]
          if len(ctable)>1:
             bins=i.get('bins','')
             if bins=='': bins=100
             bins=int(bins)

             if dmin==dmax:
                dmin=min(ctable)
                dmax=max(ctable)

             cf=i.get('cov_factor','')
             if cf=='': cf=0.5
             cf=float(cf)

             ctable.insert(0,0.0)
             ctable.append(0.0)

             try:
                density = gaussian_kde(ctable)
                xlist = np.linspace(dmin,dmax,bins)

                if cf!=-1 and cf!='':
                   density.covariance_factor = lambda:cf
                   density._compute_covariance()

                ylist=density(xlist)

                ylist5=[0.0]
                for q in ylist:
                    ylist5.append(q)
                ylist5.append(0.0)

                ylist6=np.array(ylist5)

                xlistx=argrelextrema(ylist6, np.greater)[0] # np.less for local minima

                xlistxx=[]
                for q in xlistx:
                    xlistxx.append(q-1)

                xlistx=xlistxx

             except Exception as e:
                x=format(e)
                if x.find('singular matrix')<0:
                   ck.out('CK warning: '+x+' in analyze math.variation ...')
                pass

          else:
             xlist=[ctable[0]]
             ylist=[100.0]
             xlistx=[0]

          # Convert from numpy to float
          for q in range(0, len(xlist)):
              xlist[q]=float(xlist[q])
              ylist[q]=float(ylist[q])

          if len(xlistx)>0:
             for q in xlistx:
                 xlist2.append(float(xlist[q]))
                 ylist2.append(float(ylist[q]))

             ylist2s, xlist2s = (list(t) for t in zip(*sorted(zip(ylist2, xlist2),reverse=True)))

    return {'return':0, 'xlist':xlist, 'ylist':ylist, 
                        'xlist2':xlist2, 'ylist2':ylist2,
                        'xlist2s':xlist2s, 'ylist2s':ylist2s}

##############################################################################
# analyze speedup (prepared by Anton Lokhmotov)

def speedup(i):
    """
    Input:  {
              samples1 - list of original empirical results
              samples2 - list of new empirical results (lower than original is better)
 
              (key1)   - prefix for min/max/mean in return dict
              (key2)   - prefix for min/max/mean in return dict
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              {key1}_min
              {key2}_min

              {key1}_max
              {key2}_max

              {key1}_mean
              {key2}_mean

              {key1}_var
              {key2}_var

              {key1}_delta
              {key2}_delta

              {key1}_center
              {key2}_center

              naive_speedup
              naive_speedup_min
              naive_speedup_var - always >= 1.0
            }

    """

    s1=i['samples1']
    s2=i['samples2']

    s1min=min(s1)
    s1max=max(s1)

    s2min=min(s2)
    s2max=max(s2)

    s1mean=float(sum(s1))/len(s1)
    s2mean=float(sum(s2))/len(s2)

    s1delta=s1max-s1min
    s2delta=s2max-s2min

    s1var=None
    if s1mean!=0: s1var=s1delta/s1mean

    s2var=None
    if s2mean!=0: s2var=s2delta/s2mean

    s1center=s1min+float(s1delta/2)
    s2center=s2min+float(s2delta/2)

    # naive speedups
    ns_mean=s1mean/s2mean
    ns_min=s1min/s2min

    # check keys
    k1=i.get('key1','')
    if k1=='': k1='s1'

    k2=i.get('key2','')
    if k2=='': k2='s2'

    # Normally we should simply not use speedup if variation of both
    # variables is too high and there is an overlap ...
    # We should also perform stat analysis, but as some metric
    # we temporally calculate best and worst possible speedup
    # (again it should be used more as hints)
    
    if ns_mean>ns_min: 
       nsv=ns_mean/ns_min
    else:
       nsv=ns_min/ns_mean

    # perform statistical analysis, if available 
    # and detect expected value(s) and confidence interval
#    import pandas as pd




    rr={'return':0, k1+'_min':s1min, k1+'_max':s1max,
                    k2+'_min':s2min, k2+'_max':s2max,
                    k1+'_mean':s1mean, k2+'_mean': s2mean,
                    k1+'_var':s1var, k2+'_var': s2var,
                    k1+'_delta':s1delta, k2+'_delta': s2delta,
                    k1+'_center':s1center, k2+'_center': s2delta,
                    'naive_speedup':ns_mean, 
                    'naive_speedup_min':ns_min,
                    'naive_speedup_var':nsv}

    return rr

##############################################################################
# calculating geometric mean

def geometric_mean(i):
    """
    Input:  {
              input - list of values
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              gmean        - geometric mean
            }

    """

    import math

    x=i['input']
    y=math.exp(sum(math.log(j) for j in x) / len(x))

    return {'return':0, 'gmean':y}

##############################################################################
# convert plus minus vars to user friendly strings and with proper rounding

def process_plus_minus(i):
    """
    Input:  {
              var_mean   - mean value
              var_range  - variation

              (force_round) - if 'yes', use it in %.(force_round)f
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              var_mean  - rounded
              var_range - rounded

              string
              html
              tex
            }

    """

    import math

    vm=i['var_mean']
    vr=i['var_range']

    round=0
    if vr<1:
       vr1=1/vr
       x=int(math.log10(vr1))+1
       round=x

       y=pow(10,int(x))

       vr2=int(vr*y)
       vr=vr2/y

       vm1=int(vm*y)
       vm=vm1/y
    else:
       x=int(math.log10(vr))
       y=pow(10,int(x))

       vr2=int(vr/y)
       vr=vr2*y

       vm1=int(vm/y)
       vm=vm1*y

    fr=i.get('force_round',None)
    if fr!=None and fr!='':
       round=fr

    ff='%.'+str(round)+'f'

    x1=ff % vm
    x2=ff % vr

    s=x1 +' +- '+ x2
    h=x1 +' &plusmn; '+ x2
    t=x1 +' $\pm$ '+ x2

    return {'return':0, 'var_mean':vr, 'var_range':vr, 'string':s, 'html':h, 'tex':t}



#
# Collective Knowledge (CK web service)
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin
#

import tempfile
import base64
import urllib
import cgi
import sys
import os
cfg = {}  # Will be updated by CK (meta description of this module)
work = {}  # Will be updated by CK (temporal data)
ck = None  # Will be updated by CK (initialized CK kernel)

wfe_host = ''
wfe_port = ''

# Local settings

# Import various modules while supporting both Python 2.x and 3.x
try:
    from http.server import BaseHTTPRequestHandler, HTTPServer
except:
    from BaseHTTPServer import BaseHTTPRequestHandler, HTTPServer

try:
    import urllib.parse as urlparse
except:
    import urlparse

try:
    from urllib.parse import quote as urlquote
except:
    from urllib import quote as urlquote

try:
    from urllib.parse import unquote as urlunquote
except:
    from urllib import unquote as urlunquote

# try:
#   import http.cookies as Cookie
# except:
#   import Cookie

try:
    from socketserver import ThreadingMixIn
except:
    from SocketServer import ThreadingMixIn

##############################################################################
# Initialize module


def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return': 0}

##############################################################################
# Access CK through CMD (can detach console)


def call_ck(i):
    """
    Input:  {
              Input for CK
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              (stdout)     - stdout, if available
              (stderr)     - stderr, if available
              (std)        - stdout+stderr
            }
    """

    import subprocess
    import re

    # Check action
    action = i.get('action', '')
    if action == '':
        return {'return': 1, 'error': 'action is not defined'}

    # Check that no special characters, otherwise can run any command from CMD
    if not re.match('^[A-Za-z0-9-_]*$', action):
        return {'return': 1, 'error': 'action contains illegal characters'}

    # Generate tmp file
    # suffix is important - CK will delete such file!
    fd, fn = tempfile.mkstemp(suffix='.tmp', prefix='ck-')
    os.close(fd)

    dc = i.get('detach_console', '')
    if dc == 'yes':
        i['out'] = 'con'  # If detach, output as console

    # Prepare dummy output
    rr = {'return': 0}
    rr['stdout'] = ''
    rr['stderr'] = ''

    # Save json to temporay file
    rx = ck.save_json_to_file({'json_file': fn, 'dict': i})
    if rx['return'] > 0:
        return rx

    # Prepare command line
    cmd = 'ck '+action+' @'+fn
    if dc == 'yes':
        # Check platform
        rx = ck.get_os_ck({})
        if rx['return'] > 0:
            return rx

        plat = rx['platform']

        dci = ck.cfg.get('detached_console', {}).get(plat, {})

        dcmd = dci.get('cmd', '')
        if dcmd == '':
            return {'return': 1, 'error': 'detached console is requested but cmd is not defined in kernel configuration'}

        dcmd = dcmd.replace('$#cmd#$', cmd)

        if dci.get('use_create_new_console_flag', '') == 'yes':
            process = subprocess.Popen(dcmd, stdin=None, stdout=None, stderr=None,
                                       shell=True, close_fds=True, creationflags=subprocess.CREATE_NEW_CONSOLE)
        else:
            # Will need to do the forking
            try:
                pid = os.fork()
            except OSError as e:
                return {'return': 1, 'error': 'forking detached console failed ('+format(e)+')'}

            if pid == 0:
                os.setsid()

                pid = os.fork()
                if pid != 0:
                    os._exit(0)

                try:
                    maxfd = os.sysconf("SC_OPEN_MAX")
                except (AttributeError, ValueError):
                    maxfd = 1024

                for fd in range(maxfd):
                    try:
                        os.close(fd)
                    except OSError:
                        pass

                os.open('/dev/null', os.O_RDWR)
                os.dup2(0, 1)
                os.dup2(0, 2)

                # Normally child process
                process = os.system(dcmd)
                os._exit(0)

        stdout = ck.cfg.get('detached_console_html',
                            'Console was detached ...')
        stderr = ''
    else:
        process = subprocess.Popen(
            cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        stdout, stderr = process.communicate()

    try:
        stdout = stdout.decode('utf8')
    except Exception as e:
        pass
    try:
        stderr = stderr.decode('utf8')
    except Exception as e:
        pass

    rr['std'] = stdout+stderr
    rr['stdout'] = stdout
    rr['stderr'] = stderr

    return rr

##############################################################################
# Send error to HTTP stream


def web_err(i):
    """
    Input:  {
              http - http object
              type - content type
              bin  - bytes to output
            }

    Output: {
              return - 0
            }
    """

    http = i['http']
    tp = i['type']
    bin = i['bin']

    try:
        bin = bin.decode('utf-8')
    except Exception as e:
        pass

    if tp == 'json':
        rx = ck.dumps_json({'dict': {'return': 1, 'error': bin}})
        if rx['return'] > 0:
            bin2 = rx['error'].encode('utf8')
        else:
            bin2 = rx['string'].encode('utf-8')
    elif tp == 'con':
        bin2 = bin.encode('utf8')
    else:
        bin2 = b'<html><body><pre>'+bin.encode('utf8')+b'</pre></body></html>'

    i['bin'] = bin2
    return web_out(i)

##############################################################################
# Send error to HTTP stream


def web_out(i):
    """

    Input:  {
              http       - http object
              type       - content type
              bin        - bytes to output
              (filename) - if !='', substitute filename in headers
            }

    Output: {
              return - 0
            }
    """

    http = i['http']
    bin = i['bin']

    tp = i['type']

    if tp == '' or tp == 'web':
        tp = 'html'

    tpx = cfg['content_types'].get(tp, {})
    if len(tpx) == 0:
        tp = 'unknown'
        tpx = cfg['content_types'][tp]

    fn = i.get('filename', '')

    # Output
    for k in sorted(tpx.keys()):
        v = tpx[k]
        if fn != '':
            v = v.replace('$#filename#$', fn)
        http.send_header(k, v)

    http.send_header('Access-Control-Allow-Origin', '*')
    http.send_header('Content-Length', str(len(bin)))
    http.end_headers()

    http.wfile.write(bin)

    return {'return': 0}

##############################################################################
# Process CK web service request (both GET and POST)


def process_ck_web_request(i):
    """

    Input:  {
              http - Python http object
            }

    Output: { None }
    """

    # http object
    http = i['http']

    # Parse GET variables and path
    xget = {}
    xpath = {'host': '', 'port': '', 'first': '',
             'rest': '', 'query': ''}  # May be used in the future

    xt = 'json'

    xpath['host'] = i.get('host', '')
    xpath['port'] = i.get('port', '')

    # Check GET variables
    if http.path != '':
        http.send_response(200)

        a = urlparse.urlparse(http.path)
        xp = a.path
        xr = ''

        if xp.startswith('/'):
            xp = xp[1:]

        u = xp.find('/')
        if u >= 0:
            xr = xp[u+1:]
            xp = xp[:u]

        xt = xp

        xpath['first'] = xp
        xpath['rest'] = xr
        xpath['query'] = a.query
        b = urlparse.parse_qs(a.query, keep_blank_values=True, )

        xget = {}
        for k in b:
            #           xget[k]=b[k][0]
            xget[k] = urlunquote(b[k][0])
            if sys.version_info[0] < 3:
                xget[k] = xget[k].decode('utf8')

    # Check POST
    xpost = {}
    xpost1 = {}

    try:
        headers = http.headers
        content_type = headers.get('content-type')
        ctype = ''
        if content_type != None:
            ctype, pdict = cgi.parse_header(content_type)
            # Python3 cgi.parse_multipart expects boundary to be bytes, not str.
            if sys.version_info[0] < 3 and 'boundary' in pdict:
                pdict['boundary'] = pdict['boundary'].encode()

        if ctype == 'multipart/form-data':
            if sys.version_info[0] < 3:
                xpost1 = cgi.parse_multipart(http.rfile, pdict)
            else:
                xxpost1 = cgi.FieldStorage(fp=http.rfile, headers=headers, environ={
                                           'REQUEST_METHOD': 'POST'})
                for k in xxpost1.keys():
                    xpost1[k] = [xxpost1[k].value]
        elif ctype == 'application/x-www-form-urlencoded':
            length = int(http.headers.get('content-length'))
            s = http.rfile.read(length)
            if sys.version_info[0] > 2:
                s = s.decode('utf8')
            xpost1 = cgi.parse_qs(s, keep_blank_values=1)

    except Exception as e:
        bin = b'internal CK web service error [7101] (' + \
            format(e).encode('utf8')+')'
        web_err({'http': http, 'type': xt, 'bin': bin})
        ck.out(ck.cfg['error']+bin.decode('utf8'))
        return

    # Post processing
    for k in xpost1:
        v = xpost1[k]
        if k.endswith('[]'):
            k1 = k[:-2]
            xpost[k1] = []
            for l in v:
                xpost[k1].append(urlunquote(l))
        else:
            if k != 'file_content':
                xpost[k] = urlunquote(v[0])
            else:
                xpost[k] = v[0]

        if k == 'file_content':
            fcrt = xpost1.get('file_content_record_to_tmp', '')
            if (type(fcrt) == list and len(fcrt) > 0 and fcrt[0] == 'yes') or fcrt == 'yes':
                # suffix is important - CK will delete such file!
                fd, fn = tempfile.mkstemp(suffix='.tmp', prefix='ck-')
                os.close(fd)

                f = open(fn, 'wb')
                f.write(xpost[k])
                f.close()

                xpost[k+'_uploaded'] = fn
                del(xpost[k])
                k += '_uploaded'
            else:
                import base64
                xpost[k +
                      '_base64'] = base64.urlsafe_b64encode(xpost[k]).decode('utf8')
                del(xpost[k])
                k += '_base64'

        if sys.version_info[0] < 3:
            xpost[k] = xpost[k].decode('utf8')

    # Prepare input and check if CK json present
    ii = xget
    ii.update(xpost)

    cj = ii.get('ck_json', '').strip()
    if cj != '':
        r = ck.convert_json_str_to_dict(
            {'str': cj, 'skip_quote_replacement': 'yes'})
        if r['return'] > 0:
            bin = b'internal CK web service error [7102] (' + \
                r['error'].encode('utf8')+b')'
            web_err({'http': http, 'type': xt, 'bin': bin})
            ck.out(ck.cfg['error']+bin.decode('utf8'))
            return

        del(ii['ck_json'])
        ii.update(r['dict'])

    # Misc parameters
    dc = ii.get('detach_console', '')
    act = ii.get('action', '')

    # Check output type
    if ii.get('out', '') != '':
        xt = ii['out']

    if xt == '':
        xt = 'web'

    if xt != 'json' and xt != 'con' and xt != 'web':
        web_out({'http': http,
                 'type': 'web',
                 'bin': b'Unknown CK request ('+xt.encode('utf8')+b')!'})
        return

    # Prepare temporary output file
    fd, fn = tempfile.mkstemp(prefix='ck-')
    os.close(fd)
    os.remove(fn)

    # Check output
    if dc == 'yes':
        if ck.cfg.get('forbid_detached_console', '') == 'yes':
            web_out({'http': http,
                     'type': 'web',
                     'bin': b'Detached console is forbidden!'})
            return
    else:
        ii['out_file'] = fn
        ii['web'] = 'yes'
        if xt == 'json' or xt == 'web':
            ii['out'] = 'json_file'
        # else output to console (for remote access for example)

    ii['con_encoding'] = 'utf8'

    ii['host'] = wfe_host
    ii['port'] = wfe_port

    # Execute command *********************************************************
    if act == '':
        if cfg.get('if_web_action_not_defined', '') != '' and cfg.get('if_web_module_not_defined', '') != '':
            ii['module_uoa'] = cfg['if_web_module_not_defined']
            ii['action'] = cfg['if_web_action_not_defined']

    r = call_ck(ii)

    # Process output
    if r['return'] > 0:
        if os.path.isfile(fn):
            os.remove(fn)

        bout = r['error']

        try:
            bout = bout.encode('utf-8')
        except Exception as e:
            pass

        web_err({'http': http,
                 'type': xt,
                 'bin': bout})
        return

    # If output to console or detached console
    if xt == 'con' or dc == 'yes':
        if os.path.isfile(fn):
            os.remove(fn)

        bout = r.get('std', '').encode('utf8')

        web_out({'http': http, 'type': xt, 'bin': bout})

        return

    # If json or web
    # Try to load output file
    if not os.path.isfile(fn):
        web_err({'http': http,
                 'type': xt,
                 'bin': b'Output json file was not created, see output ('+r['std'].encode('utf8')+b')!'})
        return

    r = ck.load_text_file({'text_file': fn, 'keep_as_bin': 'yes'})
    if r['return'] > 0:
        bout = r['error']

        try:
            bout = bout.encode('utf-8')
        except Exception as e:
            pass

        web_err({'http': http, 'type': xt, 'bin': bout})

        return

    bin = r['bin']

    if os.path.isfile(fn):
        os.remove(fn)

    # Process JSON output from file
    fx = ''

    if sys.version_info[0] > 2:
        bin = bin.decode('utf-8')

    ru = ck.convert_json_str_to_dict(
        {'str': bin, 'skip_quote_replacement': 'yes'})
    if ru['return'] > 0:
        bout = ru['error']

        try:
            bout = bout.encode('utf-8')
        except Exception as e:
            pass

        web_err({'http': http, 'type': xt, 'bin': bout})

        return

    rr = ru['dict']
    if rr['return'] > 0:
        bout = rr['error']

        try:
            bout = bout.encode('utf-8')
        except Exception as e:
            pass

        web_err({'http': http, 'type': xt, 'bin': bout})
        return

    # Check if file was returned
    fr = False

    if 'file_content_base64' in rr and rr.get('filename', '') != '':
        fr = True

    # Check if download
    if (xt == 'web' and fr) or (act == 'pull' and xt != 'json'):
        import base64
        x = rr.get('file_content_base64', '')

        fx = rr.get('filename', '')
        if fx == '':
            fx = ck.cfg['default_archive_name']

        # Fixing Python bug
        if sys.version_info[0] == 3 and sys.version_info[1] < 3:
            x = x.encode('utf-8')
        else:
            x = str(x)
        # convert from unicode to str since base64 works on strings
        bin = base64.urlsafe_b64decode(x)
        # should be safe in Python 2.x and 3.x

        # Process extension
        fn1, fne = os.path.splitext(fx)
        if fne.startswith('.'):
            fne = fne[1:]
        if fne != '':
            xt = fne
        else:
            xt = 'unknown'
    else:
        # Check and output html
        if rr.get('html', '') != '':
            bin = rr['html'].encode('utf-8')
        else:
            if sys.version_info[0] > 2:  # Unknown output
                bin = bin.encode('utf-8')

    web_out({'http': http, 'type': xt, 'bin': bin, 'filename': fx})

    return {'return': 0}

##############################################################################
# Class to handle requests in separate threads


class ThreadedHTTPServer(ThreadingMixIn, HTTPServer):

    """
    """

##############################################################################
# Class to handle CK web service requests


class server_handler(BaseHTTPRequestHandler):

    """
    Input:  Python http handler
    Output: None
    """

    # Process only GET
    def do_GET(self):
        process_ck_web_request({'http': self})
        return

    # Process GET and POST
    def do_POST(self):
        process_ck_web_request({'http': self})
        return

    def log_request(self, code='-', size='-'):
        self.log_message('"%s" %s %s', self.requestline, str(code), str(size))
        return

    def log_error(self, format, *args):
        self.log_message(format, *args)
        return

##############################################################################
# start web service


def start(i):
    """

    Input:  {
              (host)        - Internal web server host
              (port)        - Internal web server port

              (wfe_host)    - External web server host
              (wfe_port)    - External web server port

              (browser)     - if 'yes', open browser
              (template)    - if !='', add template
              (wcid)         - view a given entry
                or
              (cid)
              (extra_url)   - extra URL
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    # Define internal server host.
    host = ck.cfg.get('default_host', '')
    host = i.get('host', host)
    if host == '':
        host = 'localhost'  # 'localhost' if ''

    # Define external server host.
    global wfe_host
    wfe_host = i.get('wfe_host', host)

    # Define internal server port.
    port = ck.cfg.get('default_port', '')
    port = i.get('port', port)
    if port == '':
        return {'return': 1, 'error': 'web port is not defined'}

    # Define external server port.
    global wfe_port
    wfe_port = i.get('wfe_port', port)

    # Assemble URL.
    url = host+':'+port
    wfe_url = wfe_host+':'+wfe_port

    ck.out('Starting CK web service on '+url +
           ' (configured for access at '+wfe_url+') ...')
    ck.out('')

    sys.stdout.flush()

    if i.get('browser', '') == 'yes':
        rurl = 'http://'+url

        ext = ''

        if i.get('template', '') != '':
            ext = 'template='+i['template']

        cid = i.get('wcid', '')
        if cid == '':
            cid = i.get('cid', '')

        if cid != '' and cid != 'web':
            if ext != '':
                ext += '&'
            ext += 'wcid='+cid

        if i.get('extra_url', '') != '':
            if ext != '':
                ext += '&'
            ext += i['extra_url']

        if ext != '':
            rurl += '/?'+ext

        import webbrowser
        webbrowser.open(rurl)

    try:
        server = ThreadedHTTPServer((host, int(port)), server_handler)
        # Prevent issues with socket reuse
        server.allow_reuse_address = True
        server.serve_forever()
    except KeyboardInterrupt:
        ck.out('Keyboard interrupt, terminating CK web service ...')
        server.socket.close()
        return {'return': 0}
    except OSError as e:
        return {'return': 1, 'error': 'problem starting CK web service ('+format(e)+')'}

    return {'return': 0}

##############################################################################
# test web


def test(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    h = '<B>Test CK web (with unicode)</B><BR><BR>'

    r = ck.access({'action': 'load',
                   'module_uoa': 'test',
                   'data_uoa': 'unicode'})
    if r['return'] > 0:
        return r

    d = r['dict']

    for q in d['languages']:
        h += q+'<BR>'

    return {'return': 0, 'html': h}



#
# Collective Knowledge: CK-powered Caffe crowdbenchmarking (very early prototyping)
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin, Grigori.Fursin@cTuning.org, http://fursin.net
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings

line='================================================================'

ck_url='http://cknowledge.org/repo/web.php?native_action=show&native_module_uoa=program.optimization&scenario=1eb2f50d4620903e'
ck_url1='http://cknowledge.org/repo/web.php?wcid=experiment.bench.dnn:'

ffstat='ck-stat-flat-characteristics.json'
ffmin='ck-stat-flat-min.json'

form_name='wa_web_form'
onchange='document.'+form_name+'.submit();'

#hextra='<i><center>\n'
#hextra+='This is an on-going long-term project. Please check our vision [ '
#hextra+='<a href="http://doi.acm.org/10.1145/2909437.2909449">IWOCL\'16</a>, \n'
#hextra+='<a href="http://arxiv.org/abs/1506.06256">CPC\'15</a>, \n'
#hextra+='<a href="https://www.youtube.com/watch?v=Q94yWxXUMP0">YouTube</a>, \n'
#hextra+='<a href="http://ctuning.org/cm/wiki/index.php?title=CM:data:45741e3fbcf4024b:1db78910464c9d05">wiki</a> ] '
#hextra+=' and <a href="https://github.com/dividiti/ck-caffe">CK-Caffe GitHub repo</a> for more details!'
#hextra+='</center></i>\n'
#hextra+='<br>\n'

hextra='<i><center>\n'
hextra+=' [ <a href="http://cKnowledge.org/ai">Community-driven AI R&D powered by CK</a> ], '
hextra+=' [ <a href="https://github.com/ctuning/ck-caffe2">CK-Caffe2</a> / <a href="https://github.com/dividiti/ck-caffe">CK-Caffe</a> ], '
hextra+=' [ <a href="https://github.com/ctuning/ck-tensorflow">CK-TensorFlow</a> ], '
hextra+=' [ <a href="https://en.wikipedia.org/wiki/Collective_Knowledge_(software)">Wikipedia</a>, \n'
hextra+='<a href="https://www.researchgate.net/publication/304010295_Collective_Knowledge_Towards_RD_Sustainability">paper 1</a>, \n'
hextra+='<a href="https://arxiv.org/abs/1506.06256">Paper 2</a>, \n'
hextra+='<a href="https://www.youtube.com/watch?v=Q94yWxXUMP0">YouTube CK intro</a> ], \n'
hextra+=' [ <a href="http://dividiti.blogspot.fr/2017/02/we-received-test-of-time-award-for-our.html">CGO\'17 test of time award for our interdisiplinary R&D</a> ]'
hextra+='</center></i>\n'
hextra+='<br>\n'

selector=[{'name':'Type', 'key':'dnn_type'},
          {'name':'DNN engine', 'key':'dnn_engine_name'},
          {'name':'Model', 'key':'nn_type'},
          {'name':'Platform', 'key':'plat_name', 'new_line':'yes'},
          {'name':'CPU', 'key':'cpu_name'},
          {'name':'OS', 'key':'os_name', 'new_line':'yes'},
          {'name':'GPGPU', 'key':'gpgpu_name'}]

replay_clean_vars=['no_compile','host_os','device_id']
replay_clean_env_vars=['CK_CAFFE_MODEL','CK_CAFFE_MODEL_FILE','CK_ENV_MODEL_CAFFE_WEIGHTS']
  
##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}

##############################################################################
# crowdsource these experiments

def crowdsource(i):
    """
    Input:  {
              (local)               - if 'yes', local crowd-benchmarking, instead of public
              (user)                - force different user ID/email for demos

              (choices)             - force different choices to program pipeline

              (repetitions)         - statistical repetitions (default=1), for now statistical analysis is not used (TBD)
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import copy
    import os

    # Setting output
    o=i.get('out','')
    oo=''
    if o=='con': oo='con'

    quiet=i.get('quiet','')

    er=i.get('exchange_repo','')
    if er=='': er=ck.cfg['default_exchange_repo_uoa']
    esr=i.get('exchange_subrepo','')
    if esr=='': esr=ck.cfg['default_exchange_subrepo_uoa']

    if i.get('local','')=='yes': 
       er='local'
       esr=''

    la=i.get('local_autotuning','')

    repetitions=i.get('repetitions','')
    if repetitions=='': repetitions=3
    repetitions=int(repetitions)

    record='no'

    # Check if any input has . and convert to dict
    for k in list(i.keys()):
        if k.find('.')>0:
            v=i[k]

            kk='##'+k.replace('.','#')

            del(i[k])

            r=ck.set_by_flat_key({'dict':i, 'key':kk, 'value':v})
            if r['return']>0: return r

    choices=i.get('choices',{})
    env=i.get('env',{})

    if 'env' not in choices: choices['env']={}

    r=ck.merge_dicts({'dict1':choices['env'], 'dict2':copy.deepcopy(env)})
    env={}

    xchoices=copy.deepcopy(choices)

    # Get user 
    user=''

    mcfg={}
    ii={'action':'load',
        'module_uoa':'module',
        'data_uoa':cfg['module_deps']['program.optimization']}
    r=ck.access(ii)
    if r['return']==0:
       mcfg=r['dict']

       dcfg={}
       ii={'action':'load',
           'module_uoa':mcfg['module_deps']['cfg'],
           'data_uoa':mcfg['cfg_uoa']}
       r=ck.access(ii)
       if r['return']>0 and r['return']!=16: return r
       if r['return']!=16:
          dcfg=r['dict']

       user=dcfg.get('user_email','')

    # Initialize local environment for program optimization ***********************************************************
    pi=i.get('platform_info',{})
    if len(pi)==0:
       ii=copy.deepcopy(i)
       ii['action']='initialize'
       ii['module_uoa']=cfg['module_deps']['program.optimization']
       ii['data_uoa']='caffe'
       ii['exchange_repo']=er
       ii['exchange_subrepo']=esr
       ii['skip_welcome']='yes'
       ii['skip_log_wait']='yes'
       ii['crowdtuning_type']='caffe-crowd-benchmarking'
       r=ck.access(ii)
       if r['return']>0: return r

       pi=r['platform_info']
       user=r.get('user','')

    hos=pi['host_os_uoa']
    hosd=pi['host_os_dict']

    tos=pi['os_uoa']
    tosd=pi['os_dict']
    tbits=tosd.get('bits','')

    remote=tosd.get('remote','')

    tdid=pi['device_id']

    features=pi.get('features',{})

    fplat=features.get('platform',{})
    fos=features.get('os',{})
    fcpu=features.get('cpu',{})
    fgpu=features.get('gpu',{})

    plat_name=fplat.get('name','')
    plat_uid=features.get('platform_uid','')
    os_name=fos.get('name','')
    os_uid=features.get('os_uid','')
    cpu_name=fcpu.get('name','')
    if cpu_name=='': cpu_name='unknown-'+fcpu.get('cpu_abi','')
    cpu_uid=features.get('cpu_uid','')
    gpu_name=fgpu.get('name','')
    gpgpu_name=''
    sn=fos.get('serial_number','')

    # Ask for cmd
    tp=['cpu', 'cuda', 'cuda_fp16', 'opencl']

    ck.out(line)
    ck.out('Select Caffe library type:')
    ck.out('')
    r=ck.access({'action':'select_list',
                 'module_uoa':cfg['module_deps']['choice'],
                 'choices':tp})
    if r['return']>0: return r
    xtp=r['choice']

    xtp16=''
    if xtp=='cuda_fp16':
       xtp='cuda'
       xtp16='yes'

    android=False
    if 'android' in tos: android=True

    # Get extra platform features if "cuda" or "opencl"
    if android:
       run_cmd='default'
       prog_uoa='caffe-time'
    else:
       run_cmd='time_cpu'
       prog_uoa='caffe'

    gpgpu_uid=''

    if xtp=='cuda' or xtp=='opencl':
        r=ck.access({'action':'detect',
                     'module_uoa':cfg['module_deps']['platform.gpgpu'],
                     'host_os':hos,
                     'target_os':tos,
                     'device_id':tdid,
                     'type':xtp,
                     'share':'yes',
                     'exchange_repo':er,
                     'exchange_subrepo':esr})
        if r['return']>0: return r
        gfeat=r.get('features',{})
        gpgpus=gfeat.get('gpgpu',[])

        if len(gpgpus)>0:
            gpgpu_name=gpgpus[0].get('gpgpu',{}).get('name','')
            gpgpu_uid=gpgpus[0].get('gpgpu_uoa','')

        if android:
           if xtp!='opencl':
              return {'return':1, 'error':'can\'t crowdbenchmark this type of DNN engine on selected target platform'}

           run_cmd='default'
           prog_uoa='caffe-time-opencl'
        else:
           run_cmd='time_gpu'
           if xtp16=='yes': run_cmd='time_gpu_fp16'

    # Get deps from caffe program
    r=ck.access({'action':'load',
                 'module_uoa':cfg['module_deps']['program'],
                 'data_uoa':prog_uoa})
    if r['return']>0: return r

    pp=r['path']

#    lib_dep=r['dict']['run_deps']['lib-caffe']

#    deps={'lib-caffe':lib_dep}

    # Check environment for selected type
#    r=ck.access({'action':'resolve',
#                 'module_uoa':cfg['module_deps']['env'],
#                 'deps':deps,
#                 'host_os':hos,
#                 'target_os':tos,
#                 'device_id':tdid,
#                 'out':o})
#    if r['return']>0: return r
#    deps=r['deps']

    # Prepare CK pipeline for a given workload
    ii={'action':'pipeline',

        'module_uoa':cfg['module_deps']['program'],
        'data_uoa':prog_uoa,

        'host_os':hos,
        'target_os':tos,
        'device_id':tdid,

        'skip_target':'yes',

        'prepare':'yes',

        'env':env,
        'choices':choices,
#        'dependencies':deps,
        'cmd_key':run_cmd,
        'no_state_check':'yes',
        'no_compiler_description':'yes',
        'skip_info_collection':'yes',
        'skip_calibration':'yes',
        'cpu_freq':'max',
        'gpu_freq':'max',
        'env_speed':'yes',
        'energy':'no',
        'skip_print_timers':'yes',
        'generate_rnd_tmp_dir':'no',

        'out':oo}

    rr=ck.access(ii)
    if rr['return']>0: return rr

    fail=rr.get('fail','')
    if fail=='yes':
        return {'return':10, 'error':'pipeline failed ('+rr.get('fail_reason','')+')'}

    ready=rr.get('ready','')
    if ready!='yes':
        return {'return':11, 'error':'couldn\'t prepare universal CK program workflow'}

    state=rr['state']
    tmp_dir=state.get('tmp_dir','')
    if tmp_dir=='': tmp_dir='tmp' # usually when no_compile

    deps=rr['dependencies'] # resolved deps

    # Clean pipeline
    if 'ready' in rr: del(rr['ready'])
    if 'fail' in rr: del(rr['fail'])
    if 'return' in rr: del(rr['return'])

    # Prepare high-level experiment meta
    meta={'cpu_name':cpu_name,
          'os_name':os_name,
          'plat_name':plat_name,
          'gpu_name':gpu_name,
          'dnn_type':xtp,
          'gpgpu_name':gpgpu_name,
          'cmd_key':run_cmd}

    # Process deps
    xdeps={}
    xnn=''
    xblas=''
    for k in deps:
        dp=deps[k]

        ptags=dp.get('tags',[])

        puoa=dp.get('package_uoa','')
        if puoa=='':
           puoa=dp.get('cus',{}).get('used_package_uoa','')

        dname=dp.get('dict',{}).get('data_name','')

        if k=='caffemodel':
            xnn=dname

            j1=xnn.rfind('(')
            if j1>0:
                xnn=xnn[j1+1:-1]

        xdeps[k]={'name':dp.get('name',''), 'data_name':dname, 'ver':dp.get('ver',''), 'package_uoa':puoa, 'package_tags':ptags}

    # versions of engine sub deps
    dvers={}
    mdep=deps['lib-caffe']
    mdeps=mdep.get('dict',{}).get('deps',{})

    for k in mdeps:
        dvers[k]=mdeps[k].get('ver','')

    meta['xversions']=dvers
    meta['xdeps']=xdeps
    meta['nn_type']=xnn
    meta['choices']=xchoices

    mmeta=copy.deepcopy(meta)

    # Extra meta which is not used to search similar case ...
    mmeta['platform_uid']=plat_uid
    mmeta['os_uid']=os_uid
    mmeta['cpu_uid']=cpu_uid
    mmeta['gpgpu_uid']=gpgpu_uid
    mmeta['user']=user

    # Check if already exists (to aggregate stats)
    aggregated_stats={}
    rduid=''
    found=False

    if o=='con':
        ck.out('')
        ck.out('Checking if results already exists in a public repo (to aggregate statistics) ...')

    # Find remote entry
    ii={'action':'search',
        'module_uoa':work['self_module_uid'],
        'repo_uoa':er,
        'remote_repo_uoa':esr,
        'search_dict':{'meta':meta}}
    rx=ck.access(ii)
    if rx['return']>0: return rx

    lst=rx['lst']

    if len(lst)==1:
        rduid=lst[0]['data_uid']
        found=True

        if o=='con':
           ck.out('')
           ck.out('Results found. Pre-loading aggregated stats from '+rduid+' ...')

        # Load stats
        rx=ck.access({'action':'load',
                      'module_uoa':work['self_module_uid'],
                      'data_uoa':rduid,
                      'repo_uoa':er,
                      'remote_repo_uoa':esr,
                      'load_extra_json_files':[ffstat]})
        if rx['return']==0:
           aggregated_stats=rx.get('extra_json_files',{}).get(ffstat,{})
        else:
           ck.out('')
           ck.out('WARNING: couldn\'t load data ('+rx['error']+')')
    else:
       rx=ck.gen_uid({})
       if rx['return']>0: return rx
       rduid=rx['data_uid']

    # Run CK pipeline *****************************************************
    pipeline=copy.deepcopy(rr)
    if len(choices)>0:
        r=ck.merge_dicts({'dict1':pipeline['choices'], 'dict2':xchoices})
        if r['return']>0: return r

    ii={'action':'autotune',
        'module_uoa':cfg['module_deps']['pipeline'],

        'host_os':hos,
        'target_os':tos,
        'device_id':tdid,

        'iterations':1,
        'repetitions':repetitions,

        'collect_all':'yes',
        'process_multi_keys':['##characteristics#*'],

        'tmp_dir':tmp_dir,

        'pipeline':pipeline,

        'stat_flat_dict':aggregated_stats,

        "features_keys_to_process":["##choices#*"],

        "record_params": {
          "search_point_by_features":"yes"
        },

        'out':oo}

    rrr=ck.access(ii)
    if rrr['return']>0: return rrr

    ls=rrr.get('last_iteration_output',{})
    state=ls.get('state',{})
    xchoices=copy.deepcopy(ls.get('choices',{}))
    lsaf=rrr.get('last_stat_analysis',{}).get('dict_flat',{})

    real_proto=xchoices.get('env',{}).get('CK_CAFFE_MODEL','') # to push to server

    ddd={'meta':mmeta}

    ddd['choices']=xchoices

    features=ls.get('features',{})

    deps=ls.get('dependencies',{})

    fail=ls.get('fail','')
    fail_reason=ls.get('fail_reason','')

    ch=ls.get('characteristics',{})

    # Save pipeline
    ddd['state']={'fail':fail, 'fail_reason':fail_reason}
    ddd['characteristics']=ch

    ddd['user']=user

    # Add files
    ddd['file_stat']=ffstat

    if real_proto!='':
       ddd['file_model_topology']=os.path.basename(real_proto)

    if not found:
       if o=='con':
          ck.out('')
          ck.out('Saving results to the remote public repo ('+rduid+') ...')

       # Update meta
       rx=ck.access({'action':'add',
                     'module_uoa':work['self_module_uid'],
                     'data_uoa':rduid,
                     'repo_uoa':er,
                     'remote_repo_uoa':esr,
                     'dict':ddd,
                     'sort_keys':'yes'})
       if rx['return']>0: return rx

       # Push real proto
       if real_proto!='':
          if o=='con':
             ck.out('')
             ck.out('Pushing prototxt to the remote public repo ...')

          rx=ck.access({'action':'push',
                        'module_uoa':work['self_module_uid'],
                        'data_uoa':rduid,
                        'repo_uoa':er,
                        'remote_repo_uoa':esr,
                        'filename':real_proto,
                        'overwrite':'yes'})
          if rx['return']>0: return rx

    # Push statistical characteristics
    if o=='con':
       ck.out('')
       ck.out('Pushing file with statistics to server ...')

    fstat=os.path.join(pp,tmp_dir,ffstat)

    r=ck.save_json_to_file({'json_file':fstat, 'dict':lsaf, 'sort_keys':'yes'})
    if r['return']>0: return r

    rx=ck.access({'action':'push',
                  'module_uoa':work['self_module_uid'],
                  'data_uoa':rduid,
                  'repo_uoa':er,
                  'remote_repo_uoa':esr,
                  'filename':fstat,
                  'overwrite':'yes'})
    if rx['return']>0: return rx

    os.remove(fstat)

    # Info
    if o=='con':
        ck.out('')
        ck.out('Succesfully recorded results in remote repo (Entry UID='+rduid+')')
   
        # Check host URL prefix and default module/action
        url=ck_url+'&highlight_uid='+rduid
        ck.out('')
        ck.out('You can see your results at the following URL:')
        ck.out('')
        ck.out(url)

    return {'return':0}

##############################################################################
# show results

def show(i):
    """
    Input:  {
               (crowd_module_uoa) - if rendered from experiment crowdsourcing
               (crowd_key)        - add extra name to Web keys to avoid overlapping with original crowdsourcing HTML
               (crowd_on_change)  - reuse onchange doc from original crowdsourcing HTML
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import os

    st=''

    cmuoa=i.get('crowd_module_uoa','')
    ckey=i.get('crowd_key','')

    conc=i.get('crowd_on_change','')
    if conc=='':
        conc=onchange

    hi_uid=i.get('highlight_uid','')

    h=''
#    h='<hr>\n'
    h+='<center>\n'
    h+='\n\n<script language="JavaScript">function copyToClipboard (text) {window.prompt ("Copy to clipboard: Ctrl+C, Enter", text);}</script>\n\n' 

#    h+='<h2>Aggregated results from Caffe crowd-benchmarking (time, accuracy, energy, cost, ...)</h2>\n'

    h+=hextra

    # Check host URL prefix and default module/action
    rx=ck.access({'action':'form_url_prefix',
                  'module_uoa':'wfe',
                  'host':i.get('host',''), 
                  'port':i.get('port',''), 
                  'template':i.get('template','')})
    if rx['return']>0: return rx
    url0=rx['url']
    template=rx['template']

    url=url0
    action=i.get('action','')
    muoa=i.get('module_uoa','')

    st=''

    url+='action=index&module_uoa=wfe&native_action='+action+'&'+'native_module_uoa='+muoa
    url1=url

    # List entries
    ii={'action':'search',
        'module_uoa':work['self_module_uid'],
        'add_meta':'yes'}

    if cmuoa!='':
        ii['module_uoa']=cmuoa

    r=ck.access(ii)
    if r['return']>0: return r

    lst=r['lst']

    # Check unique entries
    choices={}
    wchoices={}

    for q in lst:
        d=q['meta']
        meta=d.get('meta',{})

        for kk in selector:
            kx=kk['key']
            k=ckey+kx

            if k not in choices: 
                choices[k]=[]
                wchoices[k]=[{'name':'','value':''}]

            kflat=kk.get('flat_key','')
            if kflat=='': kflat='##'+kx

            rx=ck.get_by_flat_key({'dict':meta, 'key':kflat})
            if rx['return']>0: return rx
            v=rx['value']
            if v==None: v=''

            if v!='':
                if v not in choices[k]: 
                    choices[k].append(v)
                    wchoices[k].append({'name':v, 'value':v})


    # Prepare query div ***************************************************************
    if cmuoa=='':
        # Start form + URL (even when viewing entry)
        r=ck.access({'action':'start_form',
                     'module_uoa':cfg['module_deps']['wfe'],
                     'url':url1,
                     'name':form_name})
        if r['return']>0: return r
        h+=r['html']

    for kk in selector:
        kx=kk['key']
        k=ckey+kx
        n=kk['name']

        nl=kk.get('new_line','')
        if nl=='yes':
            h+='<br>\n<div id="ck_entries_space8"></div>\n'

        v=''
        if i.get(k,'')!='':
            v=i[k]
            kk['value']=v

        # Show hardware
        ii={'action':'create_selector',
            'module_uoa':cfg['module_deps']['wfe'],
            'data':wchoices.get(k,[]),
            'name':k,
            'onchange':conc, 
            'skip_sort':'no',
            'selected_value':v}
        r=ck.access(ii)
        if r['return']>0: return r

        h+='<b>'+n+':</b> '+r['html'].strip()+'\n'

    # Check hidden
    if hi_uid!='':
        h+='<input type="hidden" name="highlight_uid" value="'+hi_uid+'">\n'

    h+='<br><br>'

    # Prune list
    plst=[]
    for q in lst:
        d=q['meta']
        meta=d.get('meta',{})

        # Check selector
        skip=False
        for kk in selector:
            k=kk['key']
            n=kk['name']
            v=kk.get('value','')

            kflat=kk.get('flat_key','')
            if kflat=='': kflat='##'+k

            rx=ck.get_by_flat_key({'dict':meta, 'key':kflat})
            if rx['return']>0: return rx
            vxx=rx['value']
            if vxx==None: vxx=''

            if v!='' and vxx!=v:
                skip=True

        if not skip:
            plst.append(q)

    # Check if too many
    lplst=len(plst)
    if lplst==0:
        h+='<b>No results found!</b>'
        return {'return':0, 'html':h, 'style':st}
    elif lplst>300:
        h+='<b>Too many entries to show ('+str(lplst)+') - please, prune list further!</b>'
        return {'return':0, 'html':h, 'style':st}

    # Prepare table
    h+='<table border="1" cellpadding="7" cellspacing="0">\n'

    ha='align="center" valign="top"'
    hb='align="left" valign="top"'

    h+='  <tr style="background-color:#dddddd">\n'
    h+='   <td '+ha+'><b>#</b></td>\n'
    h+='   <td '+ha+'><b>Platform</b></td>\n'
    h+='   <td '+ha+'><b>OS</b></td>\n'
    h+='   <td '+ha+'><b>CPU</b></td>\n'
    h+='   <td '+ha+'><b>GPGPU</b></td>\n'
    h+='   <td '+ha+'><b>Type</b></td>\n'
    h+='   <td '+ha+'><b>DNN engine</b></td>\n'
    h+='   <td '+ha+'><b>Model</b></td>\n'
    h+='   <td '+ha+'><b>Choices (env)</b></td>\n'
    h+='   <td '+ha+'><b>FWBW<br>min time</b><br><br>(exp&nbsp;time)<br>stat.&nbsp;repetitions</td>\n'
    h+='   <td '+ha+'><b>FW</b></td>\n'
    h+='   <td '+ha+'><b>BW</b></td>\n'
    h+='   <td '+ha+'><b>Per layer</b></td>\n'
    h+='   <td '+ha+'><b>HW costs</td>\n'
    h+='   <td '+ha+'><b>All usage costs (preparation, training, inference, errors, etc)</td>\n'
    h+='   <td '+ha+'><b>Model size</b></td>\n'
    h+='   <td '+ha+'><b><a href="https://github.com/dividiti/ck-caffe/blob/master/script/explore-accuracy/explore_accuracy.20160808.ipynb">Model accuracy on ImageNet</a></td>\n'
    h+='   <td '+ha+'><b>Model topology and parameters</td>\n'
    h+='   <td '+ha+'><b>Power consumption (W)<br>min&nbsp;/&nbsp;max</td>\n'
    h+='   <td '+ha+'><b>Acoustic noise (dB)<br>min&nbsp;/&nbsp;max</td>\n'
    h+='   <td '+ha+'><b>Memory usage (MB)</td>\n'
    h+='   <td '+ha+'><b>Bug detected?</b></td>\n'
    h+='   <td '+ha+'><b>User</b></td>\n'
    h+='   <td '+ha+'><b>Replay</b></td>\n'
    h+='  <tr>\n'

    # Dictionary to hold target meta
    tm={}

    ix=0
    bgraph={'0':[]} # Just for graph demo
    if hi_uid!='':
        bgraph['1']=[]

    # Load min stat
    for q in plst:
        pmin=os.path.join(q['path'],ffmin)
        dx={'##characteristics#run#time_fwbw_ms#min':1e99}

        if os.path.isfile(pmin):
           rx=ck.load_json_file({'json_file':pmin})
           if rx['return']==0:
              dx=rx['dict']

              # Fix
              x=dx.get('##characteristics#run#time_fwbw_ms#min','')
              if x==None or x=='' or x>50000: 
                 dx['##characteristics#run#time_fwbw_ms#min']=1e99
                 if q.get('meta',{}).get('state',{}).get('fail_reason','')=='':
                    q['meta']['state']['fail']='yes'
                    q['meta']['state']['fail_reason']='strange timing'

        q['min_stat']=dx

    # Sort
    splst=sorted(plst, key=lambda x: x.get('min_stat',{}).get('##characteristics#run#time_fwbw_ms#min',0))

#    splst=sorted(plst, key=lambda x: x.get('meta',{}).get('characteristics',{}).get('run',{}).get('time_fwbw_ms',0))

    for q in splst:
        ix+=1

        duid=q['data_uid']
        path=q['path']

        d=q['meta']

        # Characteristics
        # Check if has statistics
        dstat={}
        fstat=os.path.join(path,'ck-stat-flat-characteristics.json')
        if os.path.isfile(fstat):
            r=ck.load_json_file({'json_file':fstat, 'dict':dstat})
            if r['return']>0: return r
            dstat=r['dict']

        x0=dstat.get("##characteristics#run#time_fwbw_ms#min",None)

        meta=d.get('meta',{})

        choices=d.get('choices',{})
        env=choices.get('env',{})
        params=choices.get('params',{}).get('params',{})

        xdeps=meta.get('xdeps',{})

        d_model=xdeps.get('caffemodel',{})
        d_model_name=d_model.get('data_name','')
        d_model_package_uoa=d_model.get('package_uoa','')
        d_model_ver=d_model.get('ver','')

        d_engine=xdeps.get('lib-caffe',{})
        d_engine_name=d_engine.get('data_name','')
        d_engine_package_uoa=d_engine.get('package_uoa','')
        d_engine_ver=d_engine.get('ver','')

        tp=meta.get('dnn_type','')
        nn=meta.get('nn_type','')

        plat_name=meta.get('plat_name','')
        cpu_name=meta.get('cpu_name','')
        os_name=meta.get('os_name','')
        gpgpu_name=meta.get('gpgpu_name','')

        plat_uid=meta.get('platform_uid','')
        cpu_uid=meta.get('cpu_uid','')
        os_uid=meta.get('os_uid','')
        gpu_uid=meta.get('gpu_uid','')
        gpgpu_uid=meta.get('gpgpu_uid','')

        user=meta.get('user','')

        te=d.get('characteristics',{}).get('run',{})

#        bgc='afffaf'
        bgc='dfffdf'
        fail=d.get('state',{}).get('fail','')
        fail_reason=d.get('state',{}).get('fail_reason','')
        if fail=='yes':
            if fail_reason=='': fail_reason='yes'
            bgc='ffafaf'
        elif hi_uid!='' and duid==hi_uid:
            bgc='9fff9f'
#            bgraph['0'].append([ix,None])
#            bgraph['1'].append([ix,x0])

        bg=' style="background-color:#'+bgc+';"'

        h+='  <tr'+bg+'>\n'

        # Number
        h+='   <td '+ha+'><a name="'+duid+'">'+str(ix)+'</a></td>\n'

        # Platform, etc ...
        x=plat_name
        if plat_uid!='':
            x='<a href="'+url0+'&wcid='+cfg['module_deps']['platform']+':'+plat_uid+'">'+x+'</a>'
        h+='   <td '+ha+'>'+x+'</td>\n'

        x=os_name
        if os_uid!='':
            x='<a href="'+url0+'&wcid='+cfg['module_deps']['platform']+':'+os_uid+'">'+x+'</a>'
        h+='   <td '+ha+'>'+x+'</td>\n'

        x=cpu_name
        if cpu_uid!='':
            x='<a href="'+url0+'&wcid='+cfg['module_deps']['platform.cpu']+':'+cpu_uid+'">'+x+'</a>'
        h+='   <td '+ha+'>'+x+'</td>\n'

        x=gpgpu_name
        if gpgpu_uid!='':
            x='<a href="'+url0+'&wcid='+cfg['module_deps']['platform.gpgpu']+':'+gpgpu_uid+'">'+x+'</a>'
        h+='   <td '+ha+'>'+x+'</td>\n'

        # All files
        uu1=work['self_module_uid']
        if cmuoa!='': uu1=cmuoa
        uu2=str(ix)+')&nbsp;<a href="'+url0+'&wcid='+uu1+':'+duid+'">'+duid+'</a>'
        uu3='[&nbsp;<a href="'+url0+'&wcid='+uu1+':'+duid+'">See&nbsp;raw&nbsp;files</a>&nbsp;]<br><br>('+duid+')'
        uu4=uu1+':'+duid

        # Type
        h+='   <td '+ha+'>'+tp+'</a></td>\n'

        # Engine
        x=d_engine_name

        if d_engine_package_uoa!='':
           x='<a href="'+url0+'&wcid=package:'+d_engine_package_uoa+'">'+x+'</a>'

        if x!='' and d_engine_ver!='':
           x+='\n<br><br>Version&nbsp;<b>'+d_engine_ver+'</b>'

        # Versions
        ver=''
        dver=meta.get('xversions',{})
        for dx in sorted(dver):
            vx=dver[dx]
            if vx!=None and vx!='':
               ver+=dx+': '+str(dver[dx])+'\n'

        ver=ver.replace("\'","'").replace("'","\\'").replace('\"','"').replace('"',"\\'").replace('\n','\\n')
        if ver!='':
            ver='<input type="button" class="ck_small_button" onClick="alert(\''+ver+'\');" value="See versions of all deps">'

        h+='   <td '+ha+'>'+x+'<br><br>'+ver+'</td>\n'

        # Model

        x=nn

        msize=''
        mtop=''
        mtop5=''

        if d_model_package_uoa!='':
           x='<a href="'+url0+'&wcid=package:'+d_model_package_uoa+'">'+x+'</a>'

           # Load features
           rx=ck.access({'action':'load',
                         'module_uoa':'package',
                         'data_uoa':d_model_package_uoa})
           if rx['return']==0: 
              mft=rx['dict'].get('features',{})

              msize=str(mft.get('model_size_mb',''))+'&nbsp;MB'
              mtop=str(mft.get('accuracy',''))
              mtop5=str(mft.get('accuracy_top5',''))

#        if x!='' and d_model_ver!='':
#           x+='\n<br><br>Version&nbsp;<b>'+d_model_ver+'</b>'

        h+='   <td '+ha+'>'+x+'</td>\n'

        # Choices (for now env)
#        x='<table border="0" cellpadding="0" cellspacing="2">\n'
        x=''
        for k in sorted(env):
            v=env[k]
            x+=str(k)+'='+str(v)+'\n'
#            x+='<tr><td>'+str(k)+'=</td><td>'+str(v)+'</td></tr>\n'
#        x+='</table>\n'
#        x=x.replace("'","\'").replace('"',"\\'").replace('\n','\\n')
        x=x.replace("\'","'").replace("'","\\'").replace('\"','"').replace('"',"\\'").replace('\n','\\n')

        x1=''
        if x!='':
            if env.get('CK_CAFFE_BATCH_SIZE','')!='':
               x1+='Batch&nbsp;size='+env['CK_CAFFE_BATCH_SIZE']+'<br><br>\n'
            x1+='<input type="button" class="ck_small_button" onClick="alert(\''+x+'\');" value="View all">'

        h+='   <td '+ha+'>'+x1+'</td>\n'


        x=''

        # Check if has stats
        x0=dstat.get("##characteristics#run#time_fwbw_ms#min",None)
        x0e=dstat.get("##characteristics#run#time_fwbw_ms#exp",None)
        x1=dstat.get("##characteristics#run#time_fwbw_ms#center",None)
        xr=dstat.get("##characteristics#run#time_fwbw_ms#repeats",None)
        x2=dstat.get("##characteristics#run#time_fwbw_ms#halfrange",None)
        x=''
        if x0!=None:
            x='<b>'+('%.0f'%x0)+'&nbsp;ms.</b>\n'
#            x+='('+('%.0f'%x1)+'&nbsp;&PlusMinus;&nbsp;'+('%.0f'%x2)+'&nbsp;ms.)'

        if x0e!=None and x2!=None:
            x+='<br><br>('+('%.0f'%x0e)+'&nbsp;&PlusMinus;&nbsp;'+('%.0f'%x2)+'&nbsp;ms.)\n'

        if xr!=None:
            x+='<br><i>'+str(xr)+' repetitions</i>\n'

        h+='   <td '+ha+' style="background-color:#afffaf">'+x+'</td>\n'

        if fail=='yes': x0=0
        bgraph['0'].append([ix,x0])
        if fail!='yes' and x0!=None and duid!=hi_uid:
            if hi_uid!='': bgraph['1'].append([ix,None])

        x1=dstat.get("##characteristics#run#time_fw_ms#center",None)
        x2=dstat.get("##characteristics#run#time_fw_ms#halfrange",None)
        if x1!=None and x2!=None:
            x=('%.0f'%x1)+'&nbsp;&PlusMinus;&nbsp;'+('%.0f'%x2)+'&nbsp;ms.'

        h+='   <td '+ha+'>'+x+'</td>\n'

        x1=dstat.get("##characteristics#run#time_bw_ms#center",None)
        x2=dstat.get("##characteristics#run#time_bw_ms#halfrange",None)
        if x1!=None and x2!=None:
            x=('%.0f'%x1)+'&nbsp;&PlusMinus;&nbsp;'+('%.0f'%x2)+'&nbsp;ms.'

        h+='   <td '+ha+'>'+x+'</td>\n'

        # Check all characteristics
        x=''
        x5=''
        for k in sorted(te):
            v=te[k]

            kx="##characteristics#run#"+k

            kx1=dstat.get(kx+'#center',None)
            kx2=dstat.get(kx+'#halfrange',None)

            x6=''
            if type(v)==int:
                if kx1!=None and kx2!=None:
                    x6=str(kx1)+' +- '+str(kx2)
                else:
                    x6=str(v)
            elif type(v)==float:
                if kx1!=None and kx2!=None:
                    x6=('%.1f'%kx1)+' +- '+('%.1f'%kx2)
                else:
                    x6=('%.1f'%v)

            if x6!='':
                x5+=str(k)+'='+x6+'\n'

        # Also layers
        y5=''
        for j in range(0,1000):
            k1='##characteristics#run#per_layer_info@'+str(j)+'#direction#min'
            k2='##characteristics#run#per_layer_info@'+str(j)+'#label#min'
            k3='##characteristics#run#per_layer_info@'+str(j)+'#time_ms#min'
            k4='##characteristics#run#per_layer_info@'+str(j)+'#time_ms#max'
            k5='##characteristics#run#per_layer_info@'+str(j)+'#time_ms#exp_allx'

            v1=dstat.get(k1,'')
            v2=dstat.get(k2,'')
            v3=dstat.get(k3,'')
            v4=dstat.get(k4,'')
            v5=dstat.get(k5,[])

            if v1!='' and v2!='' and v3!='' and v4!='':
               v6=0
               if len(v5)>0:
                  v6=v5[0]

               xv3=''
               xv4=''
               xv5=''

               if v3!='': xv3=('%.1f'%v3)
               if v4!='': xv4=('%.1f'%v4)
               if v6!='': xv6=('%.1f'%v6)

               if y5=='': y5='Layers:\nName (direction): min time (ms.) ; expected time (ms.) ; max time (ms.)\n'

               y5+='\n'+v2+' ('+v1+'): '+xv3+';'+xv6+';'+xv4
            else:
               break

        y5=y5.replace("\'","'").replace("'","\\'").replace('\"','"').replace('"',"\\'").replace('\n','\\n')
        if y5!='':
            x+='<a href="'+ck_url1+duid+'">Stats per layer</a><br><br>\n'
            x+='<input type="button" class="ck_small_button" onClick="alert(\''+y5+'\');" value="All layers as pop-up">'

#        x5=x5.replace("'","\'").replace('"',"\\'").replace('\n','\\n')
        x5=x5.replace("\'","'").replace("'","\\'").replace('\"','"').replace('"',"\\'").replace('\n','\\n')
        if x5!='':
            x+='<br><br><input type="button" class="ck_small_button" onClick="alert(\''+x5+'\');" value="CK vars">'

        h+='   <td '+ha+'>'+x+'</td>\n'

        # Get info about platform
        hd={}
        if plat_uid!='':
           rh=ck.access({'action':'load',
                        'module_uoa':cfg['module_deps']['platform'],
                        'data_uoa':plat_uid})
           if rh['return']==0:
              hd=rh['dict']

        # Cost (take from platform meta)
        hc='-'
        if len(hd)>0:
           costs=hd.get('features',{}).get('cost',[])
           hc=''
           for c in costs:
               if hc!='': hc+='<br>\n'
               hc+='<b>'+str(c.get('price',''))+' '+c.get('currency','')+ '</b> - '+c.get('desc','')+' ('+c.get('date','')+')'

        h+='   <td '+ha+'>'+hc+'</a></td>\n'

        # TBD: all other costs
        h+='   <td '+ha+'></a></td>\n'

        # Model size
        h+='   <td '+ha+'>'+msize+'</td>\n'

        # Accuracy
        x=''

        if mtop!='' and mtop5!='':
           x=mtop+'&nbsp;/&nbsp;'+mtop5

#        if nn=='bvlc, alexnet':
#            x='0.568279&nbsp;/&nbsp;0.799501'
#        elif nn=='bvlc, googlenet':
#            x='0.689299&nbsp;/&nbsp;0.891441'
#        elif nn=='deepscale, squeezenet, 1.1':
#            x='0.583880&nbsp;/&nbsp;0.810123'
#        elif nn=='deepscale, squeezenet, 1.0':
#            x='0.576801&nbsp;/&nbsp;0.803903'

        h+='   <td '+ha+'>'+x+'</td>\n'

        # Model topology
        x=''

        fmt=d.get('file_model_topology','')
        if fmt!='':
           pfmt=os.path.join(path,fmt)
           if os.path.isfile(pfmt):
               x='<a href="'+url0+'&action=pull&common_action=yes&cid='+work['self_module_uid']+':'+duid+'&filename='+fmt+'">deploy.prototxt</a>\n'

        h+='   <td '+ha+'>'+x+'</td>\n'

        # Power consumption (TBD - real measurements)
        x='-'
        if len(hd)>0:
           power=hd.get('features',{}).get('power_consumption',{})
           if len(power)>0:
              pmin=power.get('min','')
              pmax=power.get('max','')

              x=str(pmin)+' / '+str(pmax)
        h+='   <td '+ha+'>'+x+'</a></td>\n'

        # Acoustic noise (TBD - real measurements)
        x='-'
        if len(hd)>0:
           power=hd.get('features',{}).get('acoustic_noise',{})
           if len(power)>0:
              pmin=power.get('min','')
              pmax=power.get('max','')

              x=str(pmin)+' / '+str(pmax)
        h+='   <td '+ha+'>'+x+'</a></td>\n'

        # Memory usage
        x=''

        mem=dstat.get("##characteristics#run#memory_mbytes#max",None)
        if mem!=None:
           x=str(int(mem))+' MB'

        h+='   <td '+ha+'>'+x+'</td>\n'


        # Crowdsourcing bug detection
        x=fail_reason
        if x=='': 
            x=''
        else:
            fail_reason=fail_reason.replace("\'","'").replace("'","\\'").replace('\"','"').replace('"',"\\'").replace('\n','\\n')
            x='Yes <input type="button" class="ck_small_button" onClick="alert(\''+fail_reason+'\');" value="Log">'

        h+='   <td '+ha+'>'+x+'</td>\n'


        h+='   <td '+ha+'><a href="'+url0+'&action=index&module_uoa=wfe&native_action=show&native_module_uoa=experiment.user">'+user+'</a></td>\n'

        h+='   <td '+ha+'><input type="button" class="ck_small_button" onClick="copyToClipboard(\'ck replay '+uu4+' '+ck.cfg.get('add_extra_to_replay','')+'\');" value="Replay"><br><br>\n'
        h+='              '+uu3+'</td>\n'

        h+='  <tr>\n'

    h+='</table>\n'
    h+='</center>\n'

    if cmuoa=='':
        h+='</form>\n'

    if len(bgraph['0'])>0:
       ii={'action':'plot',
           'module_uoa':cfg['module_deps']['graph'],

           "table":bgraph,

           "h_lines":[1.0],

           "ymin":0,

           "ignore_point_if_none":"yes",

           "plot_type":"d3_2d_bars",

           "display_y_error_bar":"no",

           "title":"Powered by Collective Knowledge",

           "x_ticks_period":10,

           "axis_x_desc":"Experiment",
           "axis_y_desc":"Neural network total time (ms.)",

           "plot_grid":"yes",

           "d3_div":"ck_interactive",

           "image_width":"900",
           "image_height":"400",

           "wfe_url":url0}

       r=ck.access(ii)
       if r['return']==0:
          x=r.get('html','')
          if x!='':
             st+=r.get('style','')

             h+='<br>\n'
             h+='<center>\n'
             h+='<div id="ck_box_with_shadow" style="width:920px;">\n'
             h+=' <div id="ck_interactive" style="text-align:center">\n'
             h+=x+'\n'
             h+=' </div>\n'
             h+='</div>\n'
             h+='</center>\n'

    return {'return':0, 'html':h, 'style':st}

##############################################################################
# browse public results

def browse(i):
    """
    Input:  {
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import webbrowser

    ck.out('Opening web page '+ck_url+' ...')

    webbrowser.open(ck_url)

    return {'return':0}

##############################################################################
# show info for all layers

def html_viewer(i):
    """
    Input:  {
              data_uoa - CK entry UOA to view
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import os

    duoa=i.get('data_uoa','')

    # Load entry
    r=ck.access({'action':'load',
                 'module_uoa':work['self_module_uid'],
                 'data_uoa':duoa})
    if r['return']>0: return r
    p=r['path']
    d=r['dict']

    dchars=d.get('characteristics',{})
    dchoices=d.get('choices',{})
    dmeta=d.get('meta',{})

    # Load stats
    dstat={}
    fstat=os.path.join(p,'ck-stat-flat-characteristics.json')
    if os.path.isfile(fstat):
        r=ck.load_json_file({'json_file':fstat, 'dict':dstat})
        if r['return']>0: return r
        dstat=r['dict']

    # Prepare table
    h=''
#    h+='<hr>\n'
    h+='<br>\n'
    h+='<center>\n'
    h+='<h2>DNN engine and model evaluation statistics per layer (crowd-tuning)</h2><br>\n'
    h+='</center>\n'

    xdeps=dmeta.get('xdeps',{})
    lcaffe=xdeps.get('lib-caffe',{})
    lmodel=xdeps.get('caffemodel',{})

    # Prepare extra info
    h+='<p>\n'
    h+='<table border="1" cellpadding="8" cellspacing="0">\n'

    h+=' <tr>\n'
    h+='  <td><b>DNN engine name:</b></td>\n'
    h+='  <td>'+lcaffe.get('data_name','')+'</td>\n'
    h+=' </tr>\n'

    h+=' <tr>\n'
    h+='  <td><b>DNN engine version:</b></td>\n'
    h+='  <td>'+lcaffe.get('ver','')+'</td>\n'
    h+=' </tr>\n'

    h+=' <tr>\n'
    h+='  <td><b>DNN engine type:</b></td>\n'
    h+='  <td>'+dmeta.get('dnn_type','')+'</td>\n'
    h+=' </tr>\n'

    x=''

    dx=dmeta.get('xversions',{})
    for k in sorted(dx):
        v=dx[k]
        if v!='':
           if x!='': x+='<br>\n'
           x+=k+'='+str(v)+'\n'

    h+=' <tr>\n'
    h+='  <td><b>DNN engine dependencies:</b></td>\n'
    h+='  <td>'+x+'</td>\n'
    h+=' </tr>\n'

    h+=' <tr>\n'
    h+='  <td><b>DNN model name:</b></td>\n'
    h+='  <td>'+lmodel.get('data_name','')+'</td>\n'
    h+=' </tr>\n'

    h+=' <tr>\n'
    h+='  <td><b>DNN model version:</b></td>\n'
    h+='  <td>'+lmodel.get('ver','')+'</td>\n'
    h+=' </tr>\n'

    h+=' <tr>\n'
    h+='  <td><b>Batch size:</b></td>\n'
    h+='  <td>'+dchars.get('run',{}).get('REAL_ENV_CK_CAFFE_BATCH_SIZE','')+'</td>\n'
    h+=' </tr>\n'

# TBD: Need to show min,exp,max!
#    h+=' <tr>\n'
#    h+='  <td><b>FWBW time (ms.):</b></td>\n'
#    h+='  <td>'+str(dchars.get('run',{}).get('time_bw_ms',''))+'</td>\n'
#    h+=' </tr>\n'

#    h+=' <tr>\n'
#    h+='  <td><b>FW time (ms.):</b></td>\n'
#    h+='  <td>'+str(dchars.get('run',{}).get('time_fw_ms',''))+'</td>\n'
#    h+=' </tr>\n'

#    h+=' <tr>\n'
#    h+='  <td><b>BW time (ms.):</b></td>\n'
#    h+='  <td>'+str(dchars.get('run',{}).get('time_bw_ms',''))+'</td>\n'
#    h+=' </tr>\n'

    h+=' <tr>\n'
    h+='  <td><b>Platform:</b></td>\n'
    h+='  <td>'+dmeta.get('plat_name','')+'</td>\n'
    h+=' </tr>\n'

    h+=' <tr>\n'
    h+='  <td><b>OS:</b></td>\n'
    h+='  <td>'+dmeta.get('os_name','')+'</td>\n'
    h+=' </tr>\n'

    h+=' <tr>\n'
    h+='  <td><b>CPU:</b></td>\n'
    h+='  <td>'+dmeta.get('cpu_name','')+'</td>\n'
    h+=' </tr>\n'

    h+=' <tr>\n'
    h+='  <td><b>GPU:</b></td>\n'
    h+='  <td>'+dmeta.get('gpu_name','')+'</td>\n'
    h+=' </tr>\n'


    h+=' </tr>\n'
    h+='</table>\n'

    h+='<center>\n'
    h+='<p>\n'
    h+='<table border="0" cellpadding="10" cellspacing="0">\n'

    h+=' <tr>\n'
    h+='  <td><b>Name</b></td>\n'
    h+='  <td><b>Direction</b></td>\n'
    h+='  <td align="right"><b>Min time (ms.):</b></td>\n'
    h+='  <td align="right"><b>Expected time (ms.):</b></td>\n'
    h+='  <td align="right"><b>Max time (ms.):</b></td>\n'
    h+='  <td align="right"><b>Repetitions:</b></td>\n'
    h+=' </tr>\n'

    # Detecting number of layers
    jj={}

    for j in range(0,1000):
        k3='##characteristics#run#per_layer_info@'+str(j)+'#time_ms#min'
        v3=dstat.get(k3,'')

        if v3=='': break

        jj[j]=v3

    # Sorting by min time
    if i.get('all_params',{}).get('skip_sort','')!='yes':
       jj=sorted(jj, key=lambda x: jj[x], reverse=True)

    # Also layers
    for j in jj:
        k1='##characteristics#run#per_layer_info@'+str(j)+'#direction#min'
        k2='##characteristics#run#per_layer_info@'+str(j)+'#label#min'
        k3='##characteristics#run#per_layer_info@'+str(j)+'#time_ms#min'
        k4='##characteristics#run#per_layer_info@'+str(j)+'#time_ms#max'
        k5='##characteristics#run#per_layer_info@'+str(j)+'#time_ms#exp_allx'
        k7='##characteristics#run#per_layer_info@'+str(j)+'#time_ms#repeats'

        v1=dstat.get(k1,'')
        v2=dstat.get(k2,'')
        v3=dstat.get(k3,'')
        v4=dstat.get(k4,'')
        v5=dstat.get(k5,[])
        v7=dstat.get(k7,'')

        if v1!='' and v2!='' and v3!='' and v4!='':
           v6=0
           if len(v5)>0:
              v6=v5[0]

           xv3=''
           xv4=''
           xv6=''

           if v3!='':
              if v3<0.1: xv3='0'
              else: xv3='<b>'+('%.1f'%v3)+'</b>'

           if v4!='':
              if v4<0.1: xv4='0'
              else: xv4='<b>'+('%.1f'%v4)+'</b>'

           if v6!='':
              if v6<0.1: xv6='0'
              else: xv6='<b>'+('%.1f'%v6)+'</b>'

           h+=' <tr>\n'
           h+='  <td>'+v2+'</td>\n'
           h+='  <td>'+v1+'</td>\n'
           h+='  <td align="right">'+xv3+'</td>\n'
           h+='  <td align="right">'+xv6+'</td>\n'
           h+='  <td align="right">'+xv4+'</td>\n'
           h+='  <td align="right">'+str(v7)+'</td>\n'
           h+=' </tr>\n'

    h+='</table>\n'

    h+='</center>\n'

    return {'return':0, 'html':h, 'show_top':'yes'}

##############################################################################
# replay experiment (TBD)

def replay(i):
    """
    Input:  {
              (data_uoa)
              (remote)

              (host_os)
              (target_os)
              (device_id)
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import copy
    import os

    # Setting output
    o=i.get('out','')
    oo=''
    if o=='con': oo='con'

    duoa=i.get('data_uoa','')
    remote=i.get('remote','')

    er=''
    esr=''

    if remote=='yes':
       er=i.get('exchange_repo','')
       if er=='': er=ck.cfg['default_exchange_repo_uoa']
       esr=i.get('exchange_subrepo','')
       if esr=='': esr=ck.cfg['default_exchange_subrepo_uoa']

    # Try to load info
    if o=='con':
       ck.out('Loading experiment entry ...')
       ck.out('')

    r=ck.access({'action':'load',
                 'module_uoa':work['self_module_uid'],
                 'data_uoa':duoa,
                 'repo_uoa':er,
                 'remote_repo_uoa':esr})
    if r['return']>0: return r

    d=r['dict']

    hos=i.get('host_os','')
    tos=i.get('target_os','')
    tdid=i.get('device_id','')

    # Check two main deps (engine and model)
    meta=d.get('meta',{})
    xdeps=meta.get('xdeps',{})

#    TBD: rebuild env by tags!
#
#    dnn=xdeps.get('lib-caffe',{})
#    model=xdeps.get('caffemodel',{})
#
#    pdnn=dnn.get('package_uoa','')
#    pmodel=model.get('package_uoa','')
#
#    preset_env={}
#    penv=[pdnn,pmodel]
#
#    for j in range(0, len(penv)):
#        px=''
#        py=penv[j]
#
#        if py!='':
#           # Search by package
#           r=ck.access({'action':'search',
#                        'module_uoa':cfg['module_deps']['env'],
#                        'search_dict':{'package_uoa':py}})
#           if r['return']>0: return r
#
#           l=r['lst']
#
#        if j==0: preset_env['lib-caffe']=px
#        elif j==1: preset_env['caffemodel']=px

    # Run pipeline
    choices=d.get('choices',{})
    
    # Clean various vars
    for k in replay_clean_vars:
        if k in choices:
           del(choices[k])

    if i.get('target_os','')!='' and not i['target_os'].startswith('android'):
       del(i['target_os'])

    env=choices.get('env',{})
    for k in replay_clean_env_vars:
        if k in env:
           del(env[k])
    choices['env']=env

    if hos!='': choices['host_os']=hos
    if tos!='': choices['target_os']=tos
    if tdid!='': choices['device_id']=tdid

    pipeline_data_uoa=choices['module_uoa']

    # Prepare pipeline
    ii={'action':'pipeline',
        'module_uoa':cfg['module_deps']['program'],
        'prepare':'yes',
        'choices':choices,
        'out':o}
    rr=ck.access(ii)
    if rr['return']>0: return rr

    fail=rr.get('fail','')
    if fail=='yes':
        return {'return':10, 'error':'pipeline failed ('+rr.get('fail_reason','')+')'}

    ready=rr.get('ready','')
    if ready!='yes':
        return {'return':11, 'error':'couldn\'t prepare universal CK program workflow'}

    # Run pipeline
    ii={'action':'run',
        'module_uoa':cfg['module_deps']['pipeline'],
        'data_uoa':pipeline_data_uoa,
        'pipeline':rr,
        'out':o}
    rr=ck.access(ii)
    if rr['return']>0: return rr

    fail=rr.get('fail','')
    if fail=='yes':
        return {'return':10, 'error':'pipeline failed ('+rr.get('fail_reason','')+')'}

    if o=='con':
       ck.out('')
       ck.out('Your results:')
       ck.out('')

       dstat=rr.get('last_stat_analysis',{}).get('dict_flat',{})

       x0=dstat.get("##characteristics#run#time_fwbw_ms#min",None)
       x0e=dstat.get("##characteristics#run#time_fwbw_ms#exp",None)

       if x0!=None:
            ck.out('* FWBW min: '+('%.0f'%x0)+' ms.')
       if x0e!=None:
            ck.out('* FWBW exp: '+('%.0f'%x0e)+' ms.')

       x1=dstat.get("##characteristics#run#time_fw_ms#min",None)
       x1e=dstat.get("##characteristics#run#time_fw_ms#exp",None)

       if x1!=None:
            ck.out('* FW   min: '+('%.0f'%x1)+' ms.')
       if x1e!=None:
            ck.out('* FW   exp: '+('%.0f'%x1e)+' ms.')

       x2=dstat.get("##characteristics#run#time_bw_ms#min",None)
       x2e=dstat.get("##characteristics#run#time_bw_ms#exp",None)

       if x2!=None:
            ck.out('* BW   min: '+('%.0f'%x2)+' ms.')
       if x2e!=None:
            ck.out('* BW   exp: '+('%.0f'%x2e)+' ms.')

    return {'return':0}



#
# Collective Knowledge (program - features)
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin, Grigori.Fursin@cTuning.org, http://fursin.net
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}

##############################################################################
# extract program static milepost features

def extract(i):
    """
    Input:  {
              (repo_uoa)        - repository UOA
              (data_uoa)        - program UOA (can be wildcards)

              (tags)            - tags to process specific programs

              (target_repo_uoa) - repo, where to save features - if =='', use repo_uoa


            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              dict         - final dict with key 'features'={...} 
            }

    """

    import os
    import json

    o=i.get('out','')
    oo=''
    if o=='con':oo=o

    muoa=cfg['module_deps']['program']
    duoa=i.get('data_uoa','')
    ruoa=i.get('repo_uoa','')

    truoa=i.get('target_repo_uoa','')

    tags=i.get('tags','')

    rx=ck.access({'action':'search',
                  'repo_uoa':ruoa,
                  'module_uoa':muoa,
                  'data_uoa':duoa,
                  'tags':tags})
    if rx['return']>0: return rx

    lst=rx['lst']

    feat1={}

    num=0
    for q in lst:
        num+=1

        duid=q['data_uid']
        duoa=q['data_uoa']

        xtruoa=truoa
        if truoa=='':
           xtruoa=q['repo_uid']

        if o=='con':
           ck.out(str(num)+') Processing '+duoa+' ...')

        ii={'action':'pipeline',
            'module_uoa':muoa,
            'data_uoa':duoa,
            'milepost':'yes',
            'out':oo,
            'quiet':'yes',
            'skip_info_collection':'yes',
            'no_run':'yes'}
        r=ck.access(ii)

        finish=False
        if r['return']>0:
           finish=True
           if o=='con':
              ck.out('')
              ck.out('CK WARNING: pipeline failed ('+r['error']+')')
              ck.out('')

              rx=ck.inp({'text':'Would you still like to record some extracted features (Y/n)? '})
              x=rx['string'].strip()
              if x!='n' and x!='N':
                 finish=False

        if not finish:
           feat=r.get('features',{}).get('program_static_milepost_features',{})

           if len(feat)>0:

              ddd={}

              found=False
              ry=ck.access({'action':'load',
                            'module_uoa':work['self_module_uid'],
                            'data_uoa':duid})
              if ry['return']==0: 
                 ddd=ry['dict']
                 found=True

              feat1=ddd.get('features',{}).get('program_static_milepost_features',{})
              rz=ck.merge_dicts({'dict1':feat1, 'dict2':feat})
              if rz['return']>0: return rz
              feat1=rz['dict1']

              if 'features' not in ddd: ddd['features']={}

              ddd['features']['program_static_milepost_features']=feat1

              ii={}
              ii['action']='add'
              if found: ii['action']='update'
              ii['module_uoa']=work['self_module_uid']
              ii['data_uoa']=duoa
              ii['data_uid']=duid
              ii['repo_uoa']=xtruoa
              ii['dict']=ddd
              ii['substitute']='yes'
              ry=ck.access(ii)
              if ry['return']>0: return ry

    return {'return':0, 'dict':{'features':feat1}}

##############################################################################
# Calculate similarity between programs based on features
#
# FGG: For now just simple Euclidean distance, however it has many drawbacks
# since it doesn't take into account "importance" of different features, etc
# We plan to use more SVN, decision trees, etc (see our papers) ...

def calculate_similarity(i):
    """
    Input:  {
              features1 - MILEPOST features for program 1
              features2 - MILEPOST features for program 2
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              distance     - basic Euclidean distance (can be > 1)
            }

    """

    import math

    prog1=i['features1']
    prog2=i['features2']

    # Check which feature to use for normalization
    fd=cfg['milepost_features_description']
    ftn=cfg['milepost_normalization_feature']

    # Check that in vectors
    distance=None # if error

    if ftn in prog1 and ftn in prog2:

       x1=prog1.get(ftn,None)
       x2=prog2.get(ftn,None)

       if x1!=None and x1!='' and x2!=None and x2!='':
          ftnp1=float(x1)
          ftnp2=float(x2)

          distance=0.0
          
          for q2 in fd:
              q=fd[q2]

              ftp1=float(prog1.get(q2,0))
              ftp2=float(prog2.get(q2,0))

              if q.get('use_for_euclidean_distance','')=='yes':
                 if q.get('normalized','')!='yes':
                    ftp1/=ftnp1
                    ftp2/=ftnp2

                 distance+=(ftp1-ftp2)*(ftp1-ftp2)

          if distance!=0.0:
             distance=math.sqrt(distance)

    return {'return':0, 'distance':distance}

##############################################################################
# show features in HTML

def show(i):
    """
    Input:  {
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    h='<h2>Static MILEPOST features</h2>\n'

    h+='<i>See <a href="http://cTuning.org/project-milepost">MILEPOST project webpage</a> and <a href="https://github.com/ctuning/reproduce-milepost-project">GitHub CK repository to reproduce MILEPOST project</a> for more details</i><br><br>\n'

    h+='<table class="ck_table" border="0">\n'

    # Check host URL prefix and default module/action
    rx=ck.access({'action':'form_url_prefix',
                  'module_uoa':'wfe',
                  'host':i.get('host',''), 
                  'port':i.get('port',''), 
                  'template':i.get('template','')})
    if rx['return']>0: return rx
    url0=rx['url']

    h+=' <tr style="background-color:#cfcfff;">\n'
    h+='  <td><b>\n'
    h+='   Feature number\n'
    h+='  </b></td>\n'
    h+='  <td><b>\n'
    h+='   Meaning\n'
    h+='  </b></td>\n'
    h+=' </tr>\n'

    mf=cfg['milepost_features_description']

    for ft in range(1,1000):
        sft=str(ft)
        if sft not in mf: break

        desc=mf[sft]['desc']

        if ft>56 and ft<66: desc+=' (*)'

        h+=' <tr>\n'
        h+='  <td valign="top">\n'
        h+='   ft'+sft+'\n'
        h+='  </td>\n'
        h+='  <td valign="top">\n'
        h+='   '+desc+'\n'
        h+='  </td>\n'
        h+=' </tr>\n'

    h+='</table><br>\n'

    h+='<i>* these features were added to MILEPOST feature extractor'
    h+=' by <a href="http://www.dcs.gla.ac.uk/~jsinger">Jeremy Singer</a></i><br><br>'


    return {'return':0, 'html':h}



#
# Collective Knowledge (various graphs for experiment)
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin, Grigori.Fursin@cTuning.org, http://fursin.net
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings
var_post_subgraph='subgraph'
var_post_cur_subgraph='cur_subgraph'
var_post_tmp_graph_file='graph_tmp_file'
var_post_refresh_graph='refresh_graph'
var_post_reset_graph='reset_graph'
var_post_autorefresh='graph_autorefresh'
var_post_autorefresh_time='graph_autorefresh_time'

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}

##############################################################################
# plot universal graph by flat dimensions

def plot(i):
    """

    Input:  {
              (load_table_from_file)                     - load table directly from file
                         or
              Select entries or table:
                 (repo_uoa) or (experiment_repo_uoa)     - can be wild cards
                 (remote_repo_uoa)                       - if remote access, use this as a remote repo UOA
                 (module_uoa) or (experiment_module_uoa) - can be wild cards
                 (data_uoa) or (experiment_data_uoa)     - can be wild cards

                 (repo_uoa_list)                       - list of repos to search
                 (module_uoa_list)                     - list of module to search
                 (data_uoa_list)                       - list of data to search

                 (search_dict)                         - search dict
                 (ignore_case)                         - if 'yes', ignore case when searching

                       OR 

                 table                                 - experiment table (if drawing from other functions)
                 (table_info)                          - point description


              (flat_keys_list)                      - list of flat keys to extract from points into table
                                                      (order is important: for example, for plot -> X,Y,Z)

              (flat_keys_list_separate_graphs)      - [ [keys], [keys], ...] - several graphs ...

              (sort_index)                          - (int) dimension to sort 

              (labels_for_separate_graphs)          - list of labels for separate graphs

              (flat_keys_index)                     - add all flat keys starting from this index 
              (flat_keys_index_end)                 - add all flat keys ending with this index (default #min)

              (customize_dots)                      - if 'yes' and MPL engine, customize color and size from table_info

              (out_to_file)                         - save picture or html to file, if supported 
                                                        (will be preserved and not replotted - useful to have a copy of an original image
                                                         when replotting graphs in interactive papers)

              (out_repo_uoa)                        - repo uoa where to save file (when reproducing graphs for interactive articles)
              (out_module_uoa)                      - module uoa where to save file (when reproducing graphs for interactive articles)
              (out_data_uoa)                        - data uoa where to save file (when reproducing graphs for interactive articles)

              (out_id)                              - graph ID in the created graph entry
              (out_common_meta)                     - graph common meta in the created graph entry
              (out_graph_extra_meta)                - graph extra meta (parameters and description)

              (save_table_to_json_file)             - save table to json file
              (save_info_table_to_json_file)        - save info table (mtable) to json file
              (save_table_to_csv_file)              - save table to csv file (need keys)

              (save_to_html)                        - if interactive or html-based graph, save to html
              (save_to_style)                       - if interactive or html-based graph, save to style (if needed)

              (display_x_error_bar)                 - if 'yes', display error bar on X axis (using next dim)
              (display_y_error_bar)                 - if 'yes', display error bar on Y axis (using next dim)
              (display_z_error_bar)                 - if 'yes', display error bar on Z axis (using next dim)

              Graphical parameters:
                plot_type                  - mpl_2d_scatter
                point_style                - dict, setting point style for each separate graph {"0", "1", etc}

                x_ticks_period             - (int) for bar graphs, put periodicity when to show number 

                xmin
                xmax
                ymin
                ymax

                bound_lines           - if 'yes', plot lines bounding all points ...
                 bound_style          - ':' by default
                 bound_color          - 'r' by default

                skip_colorbar         - if 'yes', do not show colorbar in heatmaps
                colorbar_pad          - (default = 0.15) - pad colorbar

                h_lines               - list of X for horizontal lines
                h_lines_style
                h_lines_color

                v_lines               - list of Y for vertical lines
                v_lines_style
                v_lines_color

                  If density graph:
                (bins)                - number of bins (int, default = 100)
                (cov_factor)          - float covariance factor

                d3_div                - div ID (ck_interactive). "body" if empty
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              (html)       - html, if HTML generator such as d3
              (style)      - style for html, if HTML generator such as d3
            }

    """

    import os
    import json

    o=i.get('out','')

    pst=i.get('point_style',{})

    otf=i.get('out_to_file','')
    otf_ruoa=i.get('out_repo_uoa','')
    otf_muoa=i.get('out_module_uoa','')
    otf_duoa=i.get('out_data_uoa','')

    xtp=i.get('x_ticks_period','')
    if xtp=='' or xtp==0: xtp=1
    if xtp!='': xtp=int(xtp)

    lsg=i.get('labels_for_separate_graphs',[])

    stjf=i.get('save_table_to_json_file','')
    sitjf=i.get('save_info_table_to_json_file','')
    stcf=i.get('save_table_to_csv_file','')

    table=i.get('table',[])
    mtable=i.get('table_info', [])

    rk=i.get('real_keys',[])

    cdots=i.get('customize_dots','')

    ltfj=i.get('load_table_from_file','')
    if ltfj!='':
       rx=ck.load_json_file({'json_file':ltfj})
       if rx['return']>0: return rx
       table=rx['dict']

    # Check if table already there
    if len(table)==0:
       # Get table from entries
       tmp_a=i.get('action','')
       tmp_mu=i.get('module_uoa','')

       i['action']='get'
       i['module_uoa']=cfg['module_deps']['experiment']

       if i.get('remote_repo_uoa','')!='':
          i['repo_uoa']=i['remote_repo_uoa']
          del(i['remote_repo_uoa'])
          if 'out' in i: del(i['out'])

       r=ck.access(i)
       if r['return']>0: return r

       table=r['table']
       mtable=r.get('mtable',{})
       rk=r['real_keys']

       merged_meta=r.get('merged_meta',{})
       pifs=r.get('plot_info_from_scenario',{})
       if len(pifs)>0:
          import copy
          ii=copy.deepcopy(i)
          i=pifs
          i.update(ii)

       i['action']=tmp_a
       i['module_uoa']=tmp_mu
    else:
       # If sort/substitute

       si=i.get('sort_index','')
       if si!='':
          rx=ck.access({'action':'sort_table', 
                        'module_uoa':cfg['module_deps']['experiment'], 
                        'table':table, 
                        'sort_index':si})
          if rx['return']>0: return rx
          table=rx['table']

       # Substitute all X with a loop (usually to sort Y and compare with predictions in scatter graphs, etc)
       if i.get('substitute_x_with_loop','')=='yes':
          rx=ck.access({'action':'substitute_x_with_loop', 
                        'module_uoa':cfg['module_deps']['experiment'], 
                        'table':table})
          if rx['return']>0: return rx
          table=rx['table']

    if len(table)==0:
       return {'return':1, 'error':'no points found'}

    # Check if out to module
    pp=''
    if otf_duoa!='':
       import os

       if otf_muoa=='': otf_muoa=work['self_module_uid']

       # Check extra params
       o_id=os.path.splitext(os.path.basename(otf))[0]

       o_cm=i.get('out_common_meta','')
       o_gem=i.get('out_graph_extra_meta',{})

       # Try to load first
       ddd={}
       xid=-1

       ii={'action':'load',
           'module_uoa':otf_muoa,
           'repo_uoa':otf_ruoa,
           'data_uoa':otf_duoa}
       rx=ck.access(ii)
       if rx['return']==0: 
          ddd=rx['dict']

          dddx=ddd.get('graphs',[])

          if o_id!='':
             for q in range(0, len(dddx)):
                 if dddx[q].get('id','')==o_id:
                    xid=q
                    break

       # Updating
       if 'graphs' not in ddd: 
          ddd['graphs']=[]

       ddd.update(o_cm)

       # Prepare graph params
       import copy
       ii=copy.deepcopy(i)
       for q in cfg['remove_keys_for_interactive_graphs']:
           if q in ii: del(ii[q])

       dddg={'params':ii}
       dddg.update(o_gem)

       if o_id!='':
          dddg['id']=o_id
       else:
          dddg['id']='default'

       if len(i.get('out_graph_meta',{}))>0:
          dddg.update(i['out_graph_meta'])

       if xid!=-1:
          ddd['graphs'][xid]=dddg
       else:
          ddd['graphs'].append(dddg)

       # Try to update this entry to be sure that we can record there, and get path
       ii={'action':'update',
           'module_uoa':otf_muoa,
           'repo_uoa':otf_ruoa,
           'data_uoa':otf_duoa,
           'dict':ddd,
           'substitute':'yes',
           'ignore_update':'yes'}
       rx=ck.access(ii)
       if rx['return']>0: return rx
       pp=rx['path']

    # Save table to JSON file, if needed
    if stjf!='':
       if pp!='':
          ppx=os.path.join(pp, stjf)
       else:
          ppx=stjf

       rx=ck.save_json_to_file({'json_file':ppx, 'dict':table})
       if rx['return']>0: return rx

    # Save info table to JSON file, if needed
    if sitjf!='':
       if pp!='':
          ppx=os.path.join(pp, sitjf)
       else:
          ppx=sitjf

       rx=ck.save_json_to_file({'json_file':ppx, 'dict':mtable})
       if rx['return']>0: return rx

    # Save table to CSV file, if needed
    if stcf!='':
       if pp!='':
          ppx=os.path.join(pp, stcf)
       else:
          ppx=stcf

       ii={'action':'convert_table_to_csv',
           'module_uoa':cfg['module_deps']['experiment'],
           'table':table,
           'keys':rk,
           'merge_multi_tables':'yes',
           'file_name':ppx}
       rx=ck.access(ii)
       if rx['return']>0: return rx

    # Prepare libraries
    pt=i.get('plot_type','')

    html=''
    style=''

    hlines=i.get('h_lines',[])
    vlines=i.get('v_lines',[])

    # Check if display error bars
    xerr=i.get('display_x_error_bar','')
    yerr=i.get('display_y_error_bar','')
    zerr=i.get('display_z_error_bar','')

    xerr2=i.get('display_x_error_bar2','') # type 2
    yerr2=i.get('display_y_error_bar2','')
    zerr2=i.get('display_z_error_bar2','')

    # Find min/max in all data and all dimensions / per sub-graph
    tmin=[]
    tmax=[]

    stmin={}
    stmax={}

    for g in table:
        gt=table[g]

        stmin[g]=[]
        stmax[g]=[]

        mgt=[]
        if g in mtable:
           mgt=mtable[g]

        xpst=pst.get(g,{})

        remove_permanent=False
        if xpst.get('remove_permanent','')=='yes':
           remove_permanent=True

        leave_only_permanent=False
        if xpst.get('leave_only_permanent','')=='yes':
           leave_only_permanent=True

        ngt=[]
#           for k in gt:
        for uindex in range(0,len(gt)):
            k=gt[uindex]

            if (remove_permanent or leave_only_permanent) and uindex<len(mgt):
               mu=mgt[uindex]

               if remove_permanent and mu.get('permanent','')=='yes':
                  continue

               if leave_only_permanent and mu.get('permanent','')!='yes':
                  continue

            ngt.append(k)

            if xpst.get('skip_from_dims','')=='yes':
               continue

            for d in range(0, len(k)):
                v=k[d]

                if len(tmin)<=d:
                   tmin.append(v)
                   tmax.append(v)
                else:
                   if v!=None and v<tmin[d]: tmin[d]=v
                   if v!=None and v>tmax[d]: tmax[d]=v 

                if len(stmin[g])<=d and v!=None:
                   stmin[g].append(v)
                   stmax[g].append(v)
                else:
                   if v!=None and v<stmin[g][d]: stmin[g][d]=v
                   if v!=None and v>stmax[g][d]: stmax[g][d]=v 

        table[g]=ngt

    ####################################################################### MPL ###
    if pt.startswith('mpl_'):

       import matplotlib as mpl

       if ck.cfg.get('use_internal_engine_for_plotting','')=='yes':
          mpl.use('agg') # if XWindows is not installed, use internal engine
       elif os.environ.get('CK_MPL_BACKEND','')!='':
          mpl.use(os.environ['CK_MPL_BACKEND'])

       import matplotlib.pyplot as plt

       fsize=int(i.get('font_size','10'))
       fweight=i.get('font_weight','normal')
       ffamily=i.get('font_family','arial')

       # Set font
       font=i.get('font',{})
       if len(font)==0:
          font = {'family':ffamily, 
                  'weight':fweight, 
                  'size': fsize}

       plt.rc('font', **font)

       # Configure graph
       gs=cfg['mpl_point_styles']

       sizex=i.get('mpl_image_size_x','')
       if sizex=='': sizex='9'

       sizey=i.get('mpl_image_size_y','')
       if sizey=='': sizey='5'

       dpi=i.get('mpl_image_dpi','')
       if dpi=='': dpi='100'

       if sizex!='' and sizey!='' and dpi!='':
          fig=plt.figure(figsize=(int(sizex),int(sizey)))
       else:
          fig=plt.figure()

       if i.get('plot_grid','')=='yes':
          plt.grid(True)

       bl=i.get('bound_lines','')

       if pt=='mpl_3d_scatter' or pt=='mpl_3d_trisurf':
          from mpl_toolkits.mplot3d import Axes3D
          sp=fig.add_subplot(111, projection='3d')
       else:
          sp=fig.add_subplot(111)

       if i.get('xscale_log','')=='yes': sp.set_xscale('log')
       if i.get('yscale_log','')=='yes': sp.set_yscale('log')
       if i.get('zscale_log','')=='yes': sp.set_zscale('log')

       # If density or heatmap, find min and max for both graphs:
       if pt=='mpl_1d_density' or pt=='mpl_1d_histogram' or pt=='mpl_2d_heatmap' or pt=='mpl_3d_scatter' or pt=='mpl_3d_trisurf':
          index=0
          if pt=='mpl_2d_heatmap': index=2

          dmean=0.0
          start=True
          dmin=0.0
          dmax=0.0
          it=0
          dt=0
          for g in table:
              gt=table[g]

              for k in gt:
                  v=k[index]

                  if v!=None and v!='':
                     if start: 
                        dmin=v
                        start=False
                     else: 
                        dmin=min(dmin, v)

                     if start: 
                        dmax=v
                        start=False
                     else: 
                        dmax=max(dmax, v)

                     it+=1
                     dt+=v

          if it!=0: dmean=dt/it

       # If heatmap, prepare colorbar
       if pt=='mpl_2d_heatmap' or pt=='mpl_3d_trisurf':
          from matplotlib import cm

          if len(i.get('color_dict',{}))>0:
             xcmap = mpl.colors.LinearSegmentedColormap('my_colormap', i['color_dict'], 1024)
          else:
             xcmap = plt.cm.get_cmap('coolwarm')

             if i.get('shifted_colormap','')=='yes':
                r=ck.load_module_from_path({'path':work['path'], 'module_code_name':'module_shifted_colormap', 'skip_init':'yes'})
                if r['return']>0: return r
                scm=r['code']

                xx_start=dmin
                if i.get('shifted_colormap_start','')!='':
                   xx_start=i['shifted_colormap_start']

                xx_stop=dmax
                if i.get('shifted_colormap_stop','')!='':
                   xx_stop=i['shifted_colormap_stop']

                xx_mid=1.0
                if i.get('shifted_colormap_mid','')!='':
                   xx_mix=i['shifted_colormap_mid']

                xcmap = scm.shiftedColorMap(xcmap, start=xx_start, stop=xx_stop, midpoint=xx_mid, name='shifted')

       # Check forced min/max for different axis
       xmin=i.get('xmin','')
       xmax=i.get('xmax','')
       ymin=i.get('ymin','')
       ymax=i.get('ymax','')
       zmin=i.get('zmin','')
       zmax=i.get('zmax','')

       if xmin!='':
          sp.set_xlim(left=float(xmin))
       if xmax!='':
          sp.set_xlim(right=float(xmax))
       if ymin!='':
          sp.set_ylim(bottom=float(ymin))
       if ymax!='':
          sp.set_ylim(top=float(ymax))
       if zmin!='':
          sp.set_zlim(bottom=float(zmin))
       if zmax!='':
          sp.set_zlim(top=float(zmax))

       # Check if invert axis
       if i.get('invert_x_axis','')=='yes': plt.gca().invert_xaxis()
       if i.get('invert_y_axis','')=='yes': plt.gca().invert_yaxis()
       if i.get('invert_z_axis','')=='yes': plt.gca().invert_zaxis()

       if pt=='mpl_2d_bars' or pt=='mpl_2d_lines':
          ind=[]
          gt=table['0']
          xt=0
          for q in gt:
              xt+=1

              if xt==xtp: 
                 v=q[0]
                 xt=0
              else: 
                 v=0

              ind.append(v)

          sp.set_xticks(ind)
          sp.set_xticklabels(ind, rotation=-20)

          width=0.9/len(table)

       # Iterate over separate graphs and add points
       s=0

       for g in sorted(table, key=int):
           gt=table[g]
           mgt=[]
           if g in mtable:
              mgt=mtable[g]

           lbl=''
           if s<len(lsg): lbl=lsg[s]

           xpst=pst.get(g,{})

           elw=int(xpst.get('elinewidth',0))
           xfmt=xpst.get('fmt','')

           xcapsize=xpst.get('capsize','')
           if xcapsize=='' or xcapsize==None: xcapsize=None
           else: xcapsize=int(xcapsize)

           cl=xpst.get('color','')
           if cl=='': cl=gs[s]['color']

           sz=xpst.get('size','')
           if sz=='': sz=gs[s]['size']

           connect_lines=xpst.get('connect_lines','')

           mrk=xpst.get('marker','')
           if mrk=='': mrk=gs[s]['marker']

           lst=xpst.get('line_style','')
           if lst=='': lst=gs[s].get('line_style', '-')

           heatmap=None

           if pt=='mpl_2d_scatter' or pt=='mpl_2d_bars' or pt=='mpl_2d_lines':
              mx=[]
              mxerr=[]
              my=[]
              myerr=[]
              mcolor=[]
              msize=[]

#              for u in gt:
              for uindex in range(0,len(gt)):
                  u=gt[uindex]
                  iu=0

                  # Check if extra info (color and size)
                  minfo={}
                  if uindex<len(mgt):
                     minfo=mgt[uindex]

                  xcl=cl
                  if minfo.get('color','')!='':
                     xcl=minfo['color']
                  mcolor.append(xcl)

                  xsz=sz
                  if minfo.get('size','')!='':
                     xsz=minfo['size']
                  msize.append(int(xsz))

                  # Check if no None
                  partial=False
                  for q in u:
                      if q==None:
                         partial=True
                         break

                  if not partial:
                     mx.append(u[iu])
                     iu+=1

                     if xerr=='yes':
                        mxerr.append(u[iu])
                        iu+=1 

                     my.append(u[iu])
                     iu+=1

                     if yerr=='yes':
                        myerr.append(u[iu])
                        iu+=1 

              if pt=='mpl_2d_bars':
                 mx1=[]
#                 mx2=[]
#                 names={}
#                 iq=0
                 for q in mx:
#                     if type(q)!=int and type(q)!=float:
#                        if q in names: q=names[q]
#                        else:
#                           names[q]=iq
#                           q=iq
#                           iq+=1
#                        mx2.append(str(q))
                     mx1.append(q+width*s)

                 if yerr=='yes':
                    sp.bar(mx1, my, width=width, edgecolor=cl, facecolor=cl, align='center', yerr=myerr, label=lbl) # , error_kw=dict(lw=2))
                 else:
                    sp.bar(mx1, my, width=width, edgecolor=cl, facecolor=cl, align='center', label=lbl)

              elif pt=='mpl_2d_lines':

                 if yerr=='yes':
                     sp.errorbar(mx, my, yerr=myerr, ls='none', c=cl, elinewidth=elw)
                 sp.plot(mx, my, c=cl, label=lbl)


              else:
                 draw_scatter=False
                 if xerr=='yes' and yerr=='yes':
                    if cdots=='yes':
                       sp.errorbar(mx, my, xerr=mxerr, yerr=myerr, ls='none', c=mcolor, elinewidth=elw, label=lbl, fmt=xfmt)
                    else:
                       sp.errorbar(mx, my, xerr=mxerr, yerr=myerr, ls='none', c=cl, elinewidth=elw, label=lbl, fmt=xfmt)
                 elif xerr=='yes' and yerr!='yes':
                    if cdots=='yes':
                       sp.errorbar(mx, my, xerr=mxerr, ls='none',  c=mcolor, elinewidth=elw, label=lbl, fmt=xfmt)
                    else:
                       sp.errorbar(mx, my, xerr=mxerr, ls='none',  c=cl, elinewidth=elw, label=lbl, fmt=xfmt)
                 elif yerr=='yes' and xerr!='yes':
                    if cdots=='yes':
                       sp.errorbar(mx, my, yerr=myerr, ls='none', c=mcolor, elinewidth=elw, label=lbl, fmt=xfmt, capsize=xcapsize)
                    else:
                       sp.errorbar(mx, my, yerr=myerr, ls='none', c=cl, elinewidth=elw, label=lbl, fmt=xfmt, capsize=xcapsize)
                 else:
                    draw_scatter=True

                 if draw_scatter or i.get('force_center_dot','')=='yes': 
                    if cdots=='yes':
                       sp.scatter(mx, my, s=msize, edgecolor=mcolor, c=mcolor, marker=mrk, label=lbl)
                    else:
                       sp.scatter(mx, my, s=int(sz), edgecolor=cl, c=cl, marker=mrk, label=lbl)

                 if connect_lines=='yes':
                    if cdots=='yes':
                       sp.plot(mx, my, c=mcolor, label=lbl)
                    else:
                       sp.plot(mx, my, c=cl, label=lbl)

                 if xpst.get('frontier','')=='yes':
                    # not optimal solution, but should work (need to sort to draw proper frontier)
                    a=[]
                    for q in range(0, len(mx)):
                        a.append([mx[q],my[q]])

                    b=sorted(a, key=lambda k: k[0])

                    if xpst.get('reuse_dims','')=='yes':
                       mx=[b[0][0]]
                       my=[tmax[1]]
                    elif xpst.get('dims_from_this_graph','')=='yes':
                       mx=[stmin[g][0]]
                       my=[stmax[g][1]]
                    else:
                       mx=[tmin[0]]
                       my=[tmax[1]]

                    for j in b:
                        mx.append(j[0])
                        my.append(j[1])

                    if xpst.get('reuse_dims','')=='yes':
                       mx.append(tmax[0])
                       my.append(b[-1][1])
                    elif xpst.get('dims_from_this_graph','')=='yes':
                       mx.append(stmax[g][0])
                       my.append(stmin[g][1])
                    else:
                       mx.append(tmax[0])
                       my.append(tmin[1])

                    sp.plot(mx, my, c=cl, linestyle=lst, label=lbl)

           elif pt=='mpl_1d_density' or pt=='mpl_1d_histogram':
              if not start: # I.e. we got non empty points
                 xbins=i.get('bins', 100)
                 xcov_factor=i.get('cov_factor', '')

                 mx=[]
                 for u in gt:
                     mx.append(u[0])

                 ii={'action':'analyze',
                     'min':dmin,
                     'max':dmax,
                     'module_uoa':cfg['module_deps']['math.variation'],
                     'bins':xbins,
                     'cov_factor':xcov_factor,
                     'characteristics_table':mx}

                 r=ck.access(ii)
                 if r['return']>0: return r

                 xs=r['xlist']
                 dxs=r['ylist']

                 pxs=r['xlist2s']
                 dpxs=r['ylist2s']

                 if pt=='mpl_1d_density':
                    sp.plot(xs,dxs, label=lbl)
                    sp.plot(pxs, dpxs, 'x', mec='r', mew=2, ms=8) #, mfc=None, mec='r', mew=2, ms=8)
                    sp.plot([dmean,dmean],[0,dpxs[0]],'g--',lw=2)
                 else:
                    plt.hist(mx, bins=xbins, normed=True, label=lbl)

           elif pt=='mpl_2d_heatmap' or pt=='mpl_3d_scatter' or pt=='mpl_3d_trisurf':
                mx=[]
                mxerr=[]
                my=[]
                myerr=[]
                mz=[]
                mzerr=[]

                for u in gt:
                    iu=0

                    # Check if no None
                    partial=False
                    for q in u:
                        if q==None:
                           partial=True
                           break

                    if not partial:
                       mx.append(u[iu])
                       iu+=1
                       if xerr=='yes':
                          mxerr.append(u[iu])
                          iu+=1 

                       my.append(u[iu])
                       iu+=1
                       if yerr=='yes':
                          myerr.append(u[iu])
                          iu+=1 

                       mz.append(u[iu])
                       iu+=1
                       if zerr=='yes':
                          mzerr.append(u[iu])
                          iu+=1 

                if pt=='mpl_2d_heatmap':
                   heatmap=sp.scatter(mx, my, c=mz, s=int(sz), marker=mrk, lw=elw, cmap=xcmap)
                elif pt=='mpl_3d_scatter':
                   heatmap=sp.scatter(mx,my,mz, c=cl, s=int(sz), marker=mrk, lw=elw)
                elif pt=='mpl_3d_trisurf':
#                   heatmap=sp.plot_trisurf(mx,my,mz,cmap=cm.coolwarm, lw=elw)
                   heatmap=sp.plot_trisurf(mx,my,mz,cmap=xcmap, lw=elw)
           s+=1
           if s>=len(gs):s=0

       # If heatmap, finish colors
       if (pt=='mpl_2d_heatmap' or pt=='mpl_3d_trisurf') and i.get('skip_colorbar','')!='yes':
          colorbar_pad=i.get('colorbar_pad','')
          if colorbar_pad=='': colorbar_pad=0.15
          colorbar_pad=float(colorbar_pad)

          plt.colorbar(heatmap, orientation=xpst.get('colorbar_orietation','horizontal'), label=xpst.get('colorbar_label',''), pad=colorbar_pad)

       # For lines
       dxmin=tmin[0]
       if xmin!='': dxmin=float(xmin)
       dxmax=tmax[0]
       if xmax!='': dxmax=float(xmax)
       dymin=tmin[1]
       if ymin!='': dymin=float(ymin)
       dymax=tmax[1]
       if ymax!='': dymax=float(ymax)

       # If bounds
       if bl=='yes':
          xbs=i.get('bound_style',':')
          xbc=i.get('bound_color','r')
          sp.plot([tmin[0],tmax[0]],[tmin[1],tmin[1]], linestyle=xbs, c=xbc)
          sp.plot([tmin[0],tmin[0]],[tmin[1],tmax[1]], linestyle=xbs, c=xbc)

       # If horizontal lines
       if len(hlines)>0:
          xbs=i.get('h_lines_style','--')
          xbc=i.get('h_lines_color','r')
          for q in hlines:
              sp.plot([dxmin,dxmax],[q,q], linestyle=xbs, c=xbc)

       # If horizontal lines
       if len(vlines)>0:
          xbs=i.get('v_lines_style','--')
          xbc=i.get('v_lines_color','r')
          for q in vlines:
              sp.plot([q,q],[dymin,dymax], linestyle=xbs, c=xbc)

       # Checking scaling
       if i.get('x_ticks_scale','')!='':
          xticks = sp.get_xticks()/float(i['x_ticks_scale'])
          sp.set_xticklabels(xticks)

       if i.get('y_ticks_scale','')!='':
          yticks = sp.get_yticks()/float(i['y_ticks_scale'])
          sp.set_yticklabels(yticks)

       if i.get('z_ticks_scale','')!='':
          zticks = sp.get_zticks()/float(i['z_ticks_scale'])
          sp.set_zticklabels(zticks)

       # Set axes labels
       xlab=i.get('axis_x_labels',[])
       if len(xlab)>0:
          ind=[]
          qq=0
          for q in xlab:
              ind.append(qq)
              qq+=1

          sp.set_xticks(ind)

          xrot=i.get('axis_x_rotation','')
          if xrot=='': sp.set_xticklabels(xlab)
          else:        sp.set_xticklabels(xlab, rotation=xrot)

       ylab=i.get('axis_y_labels',[])
       if len(ylab)>0:
          sp.set_yticks(ylab)
          sp.set_yticklabels(ylab)

       # Set axes names
       axd=i.get('axis_x_desc','')
       if axd!='': plt.xlabel(axd)

       ayd=i.get('axis_y_desc','')
       if ayd!='': plt.ylabel(ayd)

       atitle=i.get('title','')
       if atitle!='': plt.title(atitle)

#       handles, labels = plt.get_legend_handles_labels()
       plt.legend() #handles, labels)

       try:
          plt.tight_layout()
       except Exception:
          pass

       if otf=='':
          plt.show()
       else:
          if pp!='':
             ppx=os.path.join(pp, otf)
          else:
             ppx=otf

          plt.savefig(ppx)

    ####################################################################### D3 ###
    elif pt.startswith('d3_'):
       # Try to load template
       ppx=os.path.join(work['path'],'templates',pt+'.html')
       if not os.path.isfile(ppx):
          return {'return':1, 'error':'template for this graph is not found'}

       rx=ck.load_text_file({'text_file':ppx})
       if rx['return']>0: return rx
       html=rx['string']

       # Check if style is there (optional)
       ppx=os.path.join(work['path'],'templates',pt+'.style')
       if os.path.isfile(ppx):
          rx=ck.load_text_file({'text_file':ppx})
          if rx['return']>0: return rx
          style=rx['string']

       # Convert data table into JSON
       rx=ck.dumps_json({'dict':table})
       if rx['return']>0: return rx
       stable=rx['string']

       # Convert info table into JSON
       rx=ck.dumps_json({'dict':mtable})
       if rx['return']>0: return rx
       smtable=rx['string']

       # Convert point styles into JSON
       rx=ck.dumps_json({'dict':pst})
       if rx['return']>0: return rx
       spst=rx['string']

       html=html.replace('$#x_ticks_period#$',str(xtp))

       html=html.replace('$#display_x_error_bar#$',xerr)
       html=html.replace('$#display_y_error_bar#$',yerr)
       html=html.replace('$#display_z_error_bar#$',zerr)

       html=html.replace('$#display_x_error_bar2#$',xerr2)
       html=html.replace('$#display_y_error_bar2#$',yerr2)
       html=html.replace('$#display_z_error_bar2#$',zerr2)

       html=html.replace('$#cm_info_json#$',smtable)
       html=html.replace('$#cm_point_style_json#$',spst)

       html=html.replace('$#cm_data_json#$',stable)
       html=html.replace('$#cm_info_json#$',smtable)
       html=html.replace('$#cm_point_style_json#$',spst)

       html=html.replace('$#h_lines#$',json.dumps(hlines))
       html=html.replace('$#v_lines#$',json.dumps(vlines))

       # Set axes names
       axd=i.get('axis_x_desc','')
       html=html.replace('$#axis_x_desc#$', axd)

       ayd=i.get('axis_y_desc','')
       html=html.replace('$#axis_y_desc#$', ayd)

       size_x=i.get('image_width','')
       if size_x=='': size_x=600
       html=html.replace('$#ck_image_width#$', str(size_x))

       size_y=i.get('image_height','')
       if size_y=='': size_y=400
       html=html.replace('$#ck_image_height#$', str(size_y))

       xmin=i.get('xmin','')
       html=html.replace('$#ck_xmin#$', str(xmin))

       xmax=i.get('xmax','')
       html=html.replace('$#ck_xmax#$', str(xmax))

       ymin=i.get('ymin','')
       html=html.replace('$#ck_ymin#$', str(ymin))

       ymax=i.get('ymax','')
       html=html.replace('$#ck_ymax#$', str(ymax))

       # Save html to file (do not hardwire URLs)
       x=i.get('out_to_file','')
       if x!='':
          if pp!='':
             ppx=os.path.join(pp, x)
          else:
             ppx=x

          rx=ck.save_text_file({'text_file':ppx, 'string':html})
          if rx['return']>0: return rx

       # Save style to file, if needed
       x=i.get('save_to_style','')
       if x!='':
          if pp!='':
             ppx=os.path.join(pp, x)
          else:
             ppx=x

          rx=ck.save_text_file({'text_file':ppx, 'string':style})
          if rx['return']>0: return rx

       # Update URLs if needed (for example, to load .js files from CK repo)
       url0=i.get('wfe_url','')
       if url0=='': url0=ck.cfg.get('wfe_url_prefix','')

       html=html.replace('$#ck_root_url#$', url0)

       # Save working html locally to visualize without CK
       y=i.get('d3_div','')
       y1=''
       y2=''
       if y=='': 
          y='body'
       else:
          y1='<div id="'+y+'">\n\n'
          y2='\n</div>\n'
          y='div#'+y
          html=html.replace('$#ck_where#$',y)

       if i.get('save_to_html','')!='':
          x='<html>\n\n<style>\n'+style+'</style>\n\n'+'<body>\n\n'+y1+html+y2+'\n\n</body>\n</html>\n'
          x=x.replace('$#ck_where#$',y)

          rx=ck.save_text_file({'text_file':i['save_to_html'], 'string':x})
          if rx['return']>0: return rx

    else:
       return {'return':1, 'error':'this type of plot ('+pt+') is not supported'}

    return {'return':0, 'html':html, 'style':style}

##############################################################################
# Continuously updated plot

def continuous_plot(i):
    """
    Input:  {

            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    for q in range(0, 1000):
        r=plot(i)
        if r['return']>0: return r

        x=ck.inp({'text':'Press any key'})

    return {'return':0}

##############################################################################
# view entry as html

def html_viewer(i):
    """
    Input:  {
              data_uoa

              url_base
              url_pull

              url_cid

              (subgraph)

              url_pull_tmp
              tmp_data_uoa

              url_wiki

              html_share

              form_name     - current form name

              (all_params)
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    import os

    h=''
    st=''
    raw='no'
    top='no'

    duoa=i['data_uoa']
    burl=i['url_base']
    purl=i['url_pull']
    wurl=i.get('url_wiki','')

    url_cid=i.get('url_cid','')

    tpurl=i['url_pull_tmp']
    tpuoa=i['tmp_data_uoa']

    ap=i.get('all_params',{})

    ruoa=ap.get('ck_top_repo','')
    muoa=ap.get('ck_top_module','')

    if muoa=='': muoa=work['self_module_uid']

    cparams=ap.get('graph_params','') # current graph params

    hshare=i.get('html_share','')

    itype='png'

    # Check autorefresh
    ar=ap.get(var_post_autorefresh,'')
    if ar=='on':
       ap[var_post_refresh_graph]='yes'

    form_name=i['form_name']
    form_submit='document.'+form_name+'.submit();'

    art=ap.get(var_post_autorefresh_time,'')
    iart=5
    if art!='':
       try:
          iart=int(art)
       except ValueError:
          iart=5

    if ar=='on':
       h+='\n'
       h+='<script language="javascript">\n'
       h+=' <!--\n'
       h+='  setTimeout(\''+form_submit+'\','+str(iart*1000)+');\n'
       h+=' //-->\n'
       h+='</script>\n'
       h+='\n'

       # Set replotting
       jj={'action':'create_input',
           'module_uoa':cfg['module_deps']['wfe'],
           'type':'hidden', 
           'name': var_post_refresh_graph, 
           'value':'yes'}
       rx=ck.access(jj)
       if rx['return']>0: return rx
       h+=rx['html']+'\n'

    if duoa!='':
       # Load entry
       rx=ck.access({'action':'load',
                     'module_uoa':muoa,
                     'data_uoa':duoa})
       if rx['return']>0: return rx

       pp=rx['path']

       dd=rx['dict']
       duid=rx['data_uid']

       name=dd.get('name','')

       gsr=dd.get('get_shared_repo','')

       sruoa=dd.get('scripts_repo_uoa','')
       smuoa=dd.get('scripts_module_uoa','')
       sduoa=dd.get('scripts_data_uoa','')

       h+=' <span id="ck_entries1a">'+name+'</span><br>\n'
       h+=' <div id="ck_entries_space4"></div>\n'

       graphs=dd.get('graphs',[])

       # If more than one subgraph, prepare selector
       hsb=''

       igraph=0
       cgraph=0
       x=ap.get(var_post_cur_subgraph,'')
       try:
          cgraph=int(x)
       except ValueError:
          cgraph=0

#       sgraph=i.get(var_post_subgraph,'')
#       if sgraph=='':
       sgraph=ap.get(var_post_subgraph,'')
       if sgraph=='': sgraph=i.get('subgraph','')

       if len(graphs)>1:
          dx=[]
          jgraph=0
          for q in graphs:
              vid=q.get('id','')

              if sgraph=='': sgraph=vid

              if vid==sgraph: 
                 igraph=jgraph

              x=q.get('name','')
              if x=='': x=vid
              dx.append({'name':vid, 'value':vid})
              jgraph+=1

          jj={'action':'create_selector',
              'module_uoa':cfg['module_deps']['wfe'],
              'name': var_post_subgraph, 
              'onchange':form_submit,
              'data':dx,
              'selected_value':sgraph}

          rx=ck.access(jj)
          if rx['return']>0: return rx
          hsb=rx['html']+'\n'

          if igraph!=cgraph:
             ap[var_post_reset_graph]='yes'
             cgraph=igraph

          # Save current subgraph to detect change and reset ...
          jj={'action':'create_input',
              'module_uoa':cfg['module_deps']['wfe'],
              'type':'hidden', 
              'name': var_post_cur_subgraph, 
              'value':str(cgraph)}
          rx=ck.access(jj)
          if rx['return']>0: return rx
          h+=rx['html']+'\n'

       # Visualize
       gid=''
       if igraph<len(graphs):
          g=graphs[igraph]

          output=g.get('output','')

          gid=g.get('id','')
          if gid!='':
             # Get graph params
             if g.get('notes','')!='':
                h+='<i>'+g['notes']+'</i>'

             pjson=os.path.join(pp, gid+'.json')
             if not os.path.isfile(pjson): pjson=''

             pcsv=os.path.join(pp, gid+'.csv')
             if not os.path.isfile(pcsv): pcsv=''

             h+='<div id="ck_entries_space4"></div>\n'

             if hshare!='':
                h+='<div id="ck_entries_space4"></div>\n'
                h+=hshare
                h+=' <div id="ck_entries_space4"></div>\n'

             h+='<div style="text-align: right;">'

             if wurl!='':
                h+='[&nbsp;<a href="'+wurl+'">Discussion wiki (comments, reproducibility, etc.)</a>&nbsp;]'

             h+='</div>\n'

#             h+=' <hr class="ck_hr">\n'

             if hsb!='':
                h+=' <div id="ck_entries_space4"></div>\n'
                h+='<center>Select subgraph:&nbsp;'+hsb+'</center>\n'
#                h+=' <hr class="ck_hr">\n'

             # Check if interactive (html + style already prepared)
             if output=='html':
                image=gid+'.html'
             else:
                image=gid+'.'+itype

             params=g.get('params',{})

             problem_converting_json=''

             if var_post_reset_graph not in ap and cparams!='':
                rx=ck.convert_json_str_to_dict({'str':cparams, 'skip_quote_replacement':'yes'})
                if rx['return']>0:
                   problem_converting_json=rx['error']
                else:
                   params=rx['dict']

             rx=ck.dumps_json({'dict':params, 'sort_keys':'yes'})
             if rx['return']>0: return rx
             jparams=rx['string']

             # Check if need to regenerate
             problem=''
             image_orig=image
             himage=''
             if var_post_refresh_graph in ap:
                image_orig=''
                import copy

                ii=copy.deepcopy(params)
                ii['action']='plot'
                ii['module_uoa']=work['self_module_uoa']

                image=ap.get(var_post_tmp_graph_file,'')
                if image=='':
                   rx=ck.gen_tmp_file({'prefix':'tmp-', 'suffix':'.'+itype, 'remove_dir':'yes'})
                   if rx['return']>0: return rx
                   image=rx['file_name']

                ii['out_to_file']=image

                # Preset current entry params
                jj={'action':'create_input',
                    'module_uoa':cfg['module_deps']['wfe'],
                    'type':'hidden', 
                    'name': var_post_tmp_graph_file, 
                    'value':image}
                rx=ck.access(jj)
                if rx['return']>0: return rx
                h+=rx['html']+'\n'

                if ck.cfg.get('graph_tmp_repo_uoa','')!='':
                   ii['out_repo_uoa']=ck.cfg['graph_tmp_repo_uoa']

                ii['out_module_uoa']='tmp'
                ii['out_data_uoa']=tpuoa

#               (save_table_to_json_file)             - save table to json file

                rx=ck.access(ii)
                if rx['return']>0: 
                   problem=rx['error']

                purl=tpurl

             # Prepare html
             size_x=params.get('size_x','')
             size_y=params.get('size_y','')

             h+=' <table border="0" cellpadding="3" width="100%">\n'
             h+=' <tr>\n'

             extra=''
             if size_x!='': extra+='width="'+str(size_x)+'" '

             h+='  <td valign="top" '+extra+'>\n'

             h+='   <div id="ck_entries">\n'

             h+='    <b><small>Graph:</small></b>\n'
             if problem!='':
                h+='<br><br><span style="color:red;"><i>Problem: '+problem+'!</i></span><br>\n'
             else:
                if output=='html' and image!='':
                   # Check if style exists
                   pstyle=os.path.join(pp, gid+'.style')
                   if os.path.isfile(pstyle): 
                      rx=ck.load_text_file({'text_file':pstyle})
                      if rx['return']>0: return rx
                      st=rx['string']

                   # Generate UID
                   rx=ck.gen_uid({})
                   if rx['return']>0: return rx
                   uid=rx['data_uid']
                   div_with_uid='ck_interactive_'+uid

                   h+='<div id="'+div_with_uid+'">\n'

                   z=muoa+':'+duoa
                   if var_post_refresh_graph in ap:
                      z=url_cid

                   import os
                   image_st=os.path.splitext(image)[0]+'.style'

                   h+='$#ck_include_start#${"cid":"'+z+'", "where":"div#'+div_with_uid+'", "html":"'+image+'", "style":"'+image_st+'"}$#ck_include_stop#$\n'
                   h+='</div>\n'

                else:
                   if image!='':
                      if size_y!='': extra+='height="'+str(size_y)+'" '
                      himage='<img src="'+purl+image+'" '+extra+'>'
                      h+='   '+himage

             h+='   </div>\n'
             h+='  </td>\n'

             h+='  <td valign="top">\n'

             x='width:99%;'
             if size_y!='': x+='height:'+str(size_y)+'px;'

             h+='   <div id="ck_entries">\n'

             h+='    <b><small>Graph params (to customize/reproduce):</small></b>\n'

             if problem_converting_json!='':
                h+='<br><br><span style="color:red;"><i>'+problem_converting_json+'</i></span><br>\n'

             h+='   <textarea name="graph_params" style="'+x+'">\n'
             h+=jparams+'\n'
             h+='   </textarea><br>\n'

             h+='   </div>\n'

             h+='  </td>\n'

             h+=' </tr>\n'
             h+='</table>\n'

#             h+=' <hr class="ck_hr">\n'

             if g.get('skip_control','')!='yes':
                h+='<center>\n'
                h+='<button type="submit" name="'+var_post_refresh_graph+'">Replot graph</button>\n'
                h+='<button type="submit" name="'+var_post_reset_graph+'">Reset graph</button>\n'

                checked=''
                if ar=='on': checked=' checked '
                h+='&nbsp;&nbsp;&nbsp;Auto-replot graph:&nbsp;<input type="checkbox" name="'+var_post_autorefresh+'" id="'+var_post_autorefresh+'" onchange="submit()"'+checked+'>,'
                h+='&nbsp;seconds: <input type="text" name="'+var_post_autorefresh_time+'" value="'+str(iart)+'">\n'
                h+='</center>\n'

             if g.get('skip_reproduce','')!='yes':

   #             h+='<hr class="ck_hr">\n'

                h+='<center>\n'
                h+='<div id="ck_entries" style="background-color: #dfffbf;">\n'

                h+='<b>Reproducing graph:</b>\n'

                h+='<table width="99%">\n'
                h+=' <tr>\n'
                h+='  <td valign="top" align="left" width="44%">\n'

                h+='   <table border="0" cellpadding="5">\n'

                h+='    <tr>\n'
                h+='     <td valign="top" width="140"><b>Experiment entries:</b></td>\n'
                h+='     <td valign="top"><i>\n'

                duoal=params.get('data_uoa_list','')
                if len(duoal)>0:
                   h+='\n'
                   for q in duoal:
                       h+='<a href="'+burl+'wcid=experiment:'+q+'">'+q+'</a><br>\n'
                   h+='\n'

                h+='     </i></td>\n'
                h+='    </tr>\n'

                if smuoa!='' and sduoa!='':
                   cid=smuoa+':'+sduoa
                   if sruoa!='': cid=sruoa+':'+cid

                   h+='    <tr>\n'
                   h+='     <td valign="top"><b>Scripts to rebuild:</b></td>\n'
                   h+='     <td valign="top"><i>\n'
                   h+='      ck find '+cid+'<br>\n'
                   h+='      <a href="'+burl+'wcid='+cid+'">View in CK viewer</a>\n'
                   h+='     </i></td>\n'
                   h+='    </tr>\n'


                if output!='html':
                   h+='    <tr>\n'
                   h+='     <td valign="top"><b>Replay graph from CMD:</b></td>\n'
                   h+='     <td valign="top"><i>\n'
                   h+='      ck replay graph:'+duoa+' id='+gid+'\n'
                   h+='     </i></td>\n'
                   h+='    </tr>\n'

                h+='   </table>\n'

                h+='  </td>\n'
                h+='  <td valign="top" align="left" width="56%">\n'

                h+='   <table border="0" cellpadding="5">\n'

                if gsr!='':
                   h+='    <tr>\n'
                   h+='     <td valign="top" width="250"><b>Obtain shared CK repo with all artifacts:</b></td>\n'
                   h+='     <td valign="top">\n'
                   h+='      <i>'+gsr+'</i>\n'
                   h+='     </td>\n'
                   h+='    </tr>\n'

                refresh=False
                if var_post_refresh_graph in ap: refresh=True

   #             if (pjson!='' or pcsv!='') and image_orig!='' and himage!='':
   #             if (pjson!='' or pcsv!='') and not refresh: # if refresh, table may change
                x1=purl+gid+'.json'
                x2=purl+gid+'.csv'

                h+='    <tr>\n'
                h+='     <td valign="top"><b>Original experiment table:</b></td>\n'
                h+='     <td valign="top"><i>\n'
                if pjson!='':
                   h+='      <a href="'+x1+'">Download in JSON</a>;&nbsp;&nbsp'
                if pcsv!='':
                   h+='      <a href="'+x2+'">Download in CSV</a>\n'
                h+='     </i></td>\n'
                h+='    </tr>\n'

                if image_orig!='' and himage!='':
                   h+='    <tr>\n'
                   h+='     <td valign="top"><b>Embedd original image into interactive report/paper:</b></td>\n'
                   h+='     <td valign="top"><i>\n'
                   h+='      '+himage.replace('<','&lt;').replace('>','&gt;')+'\n'
                   h+='     </i></td>\n'
                   h+='    </tr>\n'

                h+='   </table>\n'

                h+='  </td>\n'
                h+=' </tr>\n'
                h+='</table>\n'

                h+='</div>\n'

                h+='</center>\n'

    return {'return':0, 'raw':raw, 'show_top':top, 'html':h, 'style':st}

##############################################################################
# replaying saved graph from CMD

def replay(i):
    """
    Input:  {
              (repo_uoa)   - repo UOA of s saved graph
              data_uoa     - data UOA of a saved graph
              (id)         - subgraph id
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    o=i.get('out','')

    duoa=i['data_uoa']
    ruoa=i.get('repo_uoa','')

    ii={'action':'load',
        'module_uoa':work['self_module_uid'],
        'data_uoa':duoa,
        'repo_uoa':ruoa}
    rx=ck.access(ii)
    if rx['return']>0: return rx

    dd=rx['dict']

    graphs=dd.get('graphs',[])

    if len(graphs)==0:
       return {'return':1, 'error':'no saved graphs found'}

    igraph=-1

    if len(graphs)==1:
       igraph=0
    elif len(graphs)>1:
       gid=i.get('id','')
       if gid=='':
          if o=='con':
             ck.out('Available subgraphs ID:')
             for q in graphs:
                 ck.out('  '+q.get('id',''))
             ck.out('')

          return {'return':1, 'error':'more than one subgraph found - please, specify id'}

       jgraph=0
       for q in graphs:
           if q.get('id','')==gid:
              igraph=jgraph
              break
           jgraph+=1

    if igraph==-1:
       return {'return':1, 'error':'can\'t find subgraph'}

    params=graphs[igraph].get('params',{})

    # Replaying
    params['action']='plot'
    params['module_uoa']=work['self_module_uid']

    return ck.access(params)



#
# Collective Knowledge (check program output)
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin, Grigori.Fursin@cTuning.org, http://fursin.net
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}

##############################################################################
# check numerical program output with threshold

def check_numerical(i):
    """
    Input:  {
              file1 - file1 to check
                or
              dict1

              file2 - file2 to check
                or
              dict2

              (abs_threshold) - 0 by default
            }

    Output: {
              return        - return code =  0, if successful
                                          >  0, if error
              (error)       - error text if return > 0

              failed        - if True, failed
              fail_reason   - failure reason
            }

    """

    abs_threshold=i.get('abs_threshold','')
    if abs_threshold=='': abs_threshold='0.0'
    abs_threshold=float(abs_threshold)

    dict1={}

    file1=i.get('file1','')
    if file1!='':
       r=ck.load_json_file({'json_file':file1})
       if r['return']>0: return r
       dict1=r['dict']

    if len(dict1)==0:
       return {'return':1, 'error':'dict1 to compare is empty'}

    dict2={}
    file2=i.get('file2','')
    if file2!='':
       r=ck.load_json_file({'json_file':file2})
       if r['return']>0: return r
       dict2=r['dict']

    if len(dict2)==0:
       return {'return':1, 'error':'dict2 to compare is empty'}

    # For now expect that list - later can extend it
    if type(dict1)!=list:
       return {'return':1, 'error':'non lists are not supported yet - TBD!'}

    l1=len(dict1)
    l2=len(dict2)

    if l1!=l2:
       return {'return':0, 'failed':True, 'fail_reason':'different length of numerical outputs ('+str(l1)+' vs '+str(l2)+')'}

    err=''
    for k in range(0, l1):
        v1=dict1[k]
        v2=dict2[k]

        dv=abs(v1-v2)

        if dv>abs_threshold:
           err+=str(k)+') '+str(v1)+' vs '+str(v2)+'\n'

    if err!='':
       return {'return':0, 'failed':True, 'fail_reason':'Numerical outputs differ:\n'+err}

    return {'return':0, 'failed':False}
